#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 19 17:16:23 2019

@author: ljia
"""
import numpy as np
from sklearn.manifold import TSNE, Isomap
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import zoomed_inset_axes, mark_inset
from tqdm import tqdm

from gklearn.utils.graphfiles import loadDataset, loadGXL
from gklearn.preimage.utils import kernel_distance_matrix, compute_kernel, dis_gstar, get_same_item_indices


def visualize_graph_dataset(dis_measure, visual_method, draw_figure, 
                            draw_params={}, dis_mat=None, Gn=None, 
                            median_set=None):
    
    
    def draw_zoomed_axes(Gn_embedded, ax):
        margin = 0.01
        if dis_measure == 'graph-kernel':
            index = -2
        elif dis_measure == 'ged':
            index = -1
        x1 = np.min(Gn_embedded[median_set + [index], 0]) - margin * np.max(Gn_embedded)
        x2 = np.max(Gn_embedded[median_set + [index], 0]) + margin * np.max(Gn_embedded)
        y1 = np.min(Gn_embedded[median_set + [index], 1]) - margin * np.max(Gn_embedded)
        y2 = np.max(Gn_embedded[median_set + [index], 1]) + margin * np.max(Gn_embedded)
        if (x1 < 0 and y1 < 0) or ((x1 > 0 and y1 > 0)):
            loc = 2
        else:
            loc = 3
        axins = zoomed_inset_axes(ax, 4, loc=loc) # zoom-factor: 2.5, location: upper-left
        draw_figure(axins, Gn_embedded, dis_measure=dis_measure, 
                    median_set=median_set, **draw_params)
        axins.set_xlim(x1, x2) # apply the x-limits
        axins.set_ylim(y1, y2) # apply the y-limits
        plt.yticks(visible=False)
        plt.xticks(visible=False)
        loc1 = 1 if loc == 2 else 3
        mark_inset(ax, axins, loc1=2, loc2=4, fc="none", ec="0.5")  
        
        
    if dis_mat is None:
        if dis_measure == 'graph-kernel':
            gkernel = 'untilhpathkernel'
            node_label = 'atom'
            edge_label = 'bond_type'
            dis_mat, _, _, _ = kernel_distance_matrix(Gn, node_label, edge_label, 
                                                      Kmatrix=None, gkernel=gkernel)
        elif dis_measure == 'ged':
            pass
        
    if visual_method == 'tsne':
        Gn_embedded = TSNE(n_components=2, metric='precomputed').fit_transform(dis_mat)
    elif visual_method == 'isomap':
        Gn_embedded = Isomap(n_components=2, metric='precomputed').fit_transform(dis_mat)
    print(Gn_embedded.shape)
    fig, ax = plt.subplots()
    draw_figure(plt, Gn_embedded, dis_measure=dis_measure, legend=True, 
                median_set=median_set, **draw_params)        
#    draw_zoomed_axes(Gn_embedded, ax)
    plt.show()
    plt.clf()
    
    return


def draw_figure(ax, Gn_embedded, dis_measure=None, y_idx=None, legend=False,
                median_set=None):
    from matplotlib import colors as mcolors
    colors = list(dict(mcolors.BASE_COLORS, **mcolors.CSS4_COLORS))
#    colors = ['#08306b', '#08519c', '#2171b5', '#4292c6', '#6baed6', '#9ecae1',
#              '#c6dbef', '#deebf7']
#    for i, values in enumerate(y_idx.values()):
#        for item in values:
##            ax.scatter(Gn_embedded[item,0], Gn_embedded[item,1], c=colors[i]) # , c='b')        
#            ax.scatter(Gn_embedded[item,0], Gn_embedded[item,1], c='b')
#    ax.scatter(Gn_embedded[:,0], Gn_embedded[:,1], c='b')        
    h1 = ax.scatter(Gn_embedded[median_set, 0], Gn_embedded[median_set, 1], c='b')
    if dis_measure == 'graph-kernel':
        h2 = ax.scatter(Gn_embedded[-1, 0], Gn_embedded[-1, 1], c='darkorchid') # \psi
        h3 = ax.scatter(Gn_embedded[-2, 0], Gn_embedded[-2, 1], c='gold') # gen median
        h4 = ax.scatter(Gn_embedded[-3, 0], Gn_embedded[-3, 1], c='r') #c='g', marker='+') # set median
    elif dis_measure == 'ged':
        h3 = ax.scatter(Gn_embedded[-1, 0], Gn_embedded[-1, 1], c='gold') # gen median
        h4 = ax.scatter(Gn_embedded[-2, 0], Gn_embedded[-2, 1], c='r') #c='g', marker='+') # set median        
    if legend:
#    fig.subplots_adjust(bottom=0.17)
        if dis_measure == 'graph-kernel':
            ax.legend([h1, h2, h3, h4], 
                      ['k closest graphs', 'true median', 'gen median', 'set median'])
        elif dis_measure == 'ged':       
            ax.legend([h1, h3, h4], ['k closest graphs', 'gen median', 'set median'])
#    fig.legend(handles, labels, loc='lower center', ncol=2, frameon=False) # , ncol=5, labelspacing=0.1, handletextpad=0.4, columnspacing=0.6)
#    plt.savefig('symbolic_and_non_comparison_vertical_short.eps', format='eps', dpi=300, transparent=True,
#            bbox_inches='tight')
#    plt.show()
            
    
###############################################################################
    
def visualize_distances_in_kernel():
    
    ds = {'name': 'monoterpenoides', 
          'dataset': '../datasets/monoterpenoides/dataset_10+.ds'}  # node/edge symb
    Gn, y_all = loadDataset(ds['dataset'])
#    Gn = Gn[0:50]
    fname_medians = 'expert.treelet'
    # add set median.
    fname_sm = 'results/test_k_closest_graphs/set_median.' + fname_medians + '.gxl'
    set_median = loadGXL(fname_sm)
    Gn.append(set_median)
    # add generalized median (estimated pre-image.)
    fname_gm = 'results/test_k_closest_graphs/gen_median.' + fname_medians + '.gxl'
    gen_median = loadGXL(fname_gm)
    Gn.append(gen_median)
    
    # compute distance matrix
    median_set = [22, 29, 54, 74]
    gkernel = 'treeletkernel'
    node_label = 'atom'
    edge_label = 'bond_type'
    Gn_median_set = [Gn[i].copy() for i in median_set]
    Kmatrix_median = compute_kernel(Gn + Gn_median_set, gkernel, node_label, 
                                    edge_label, True)
    Kmatrix = Kmatrix_median[0:len(Gn), 0:len(Gn)]
    dis_mat, _, _, _ = kernel_distance_matrix(Gn, node_label, edge_label, 
                                              Kmatrix=Kmatrix, gkernel=gkernel)
    print('average distances: ', np.mean(np.mean(dis_mat[0:len(Gn)-2, 0:len(Gn)-2])))
    print('min distances: ', np.min(np.min(dis_mat[0:len(Gn)-2, 0:len(Gn)-2])))
    print('max distances: ', np.max(np.max(dis_mat[0:len(Gn)-2, 0:len(Gn)-2])))

    # add distances for the image of exact median \psi.
    dis_k_median_list = []
    for idx, g in enumerate(Gn):
        dis_k_median_list.append(dis_gstar(idx, range(len(Gn), len(Gn) + len(Gn_median_set)), 
                                           [1 / len(Gn_median_set)] * len(Gn_median_set),
                                           Kmatrix_median, withterm3=False))
    dis_mat_median = np.zeros((len(Gn) + 1, len(Gn) + 1))
    for i in range(len(Gn)):
        for j in range(i, len(Gn)):
            dis_mat_median[i, j] = dis_mat[i, j]
            dis_mat_median[j, i] = dis_mat_median[i, j]
    for i in range(len(Gn)):
        dis_mat_median[i, -1] = dis_k_median_list[i]
        dis_mat_median[-1, i] = dis_k_median_list[i]
    
    # get indices by classes.
    y_idx = get_same_item_indices(y_all)
    
    # visualization.
#    visualize_graph_dataset('graph-kernel', 'tsne', Gn)
#    visualize_graph_dataset('graph-kernel', 'tsne', draw_figure, 
#                            draw_params={'y_idx': y_idx}, dis_mat=dis_mat_median)
    visualize_graph_dataset('graph-kernel', 'tsne', draw_figure, 
                            draw_params={'y_idx': y_idx}, dis_mat=dis_mat_median,
                            median_set=median_set)
        
    
def visualize_distances_in_ged():
    from gklearn.preimage.fitDistance import compute_geds
    from gklearn.preimage.ged import GED
    ds = {'name': 'monoterpenoides', 
          'dataset': '../datasets/monoterpenoides/dataset_10+.ds'}  # node/edge symb
    Gn, y_all = loadDataset(ds['dataset'])
#    Gn = Gn[0:50]
    # add set median.
    fname_medians = 'expert.treelet'
    fname_sm = 'preimage/results/test_k_closest_graphs/set_median.' + fname_medians + '.gxl'
    set_median = loadGXL(fname_sm)
    Gn.append(set_median)
    # add generalized median (estimated pre-image.)
    fname_gm = 'preimage/results/test_k_closest_graphs/gen_median.' + fname_medians + '.gxl'
    gen_median = loadGXL(fname_gm)
    Gn.append(gen_median)
    
    # compute/load ged matrix.
#    # compute.
##    k = 4
##    edit_costs = [0.16229209837639536, 0.06612870523413916, 0.04030113378793905, 0.20723547009415202, 0.3338607220394598, 0.27054392518077297]
#    edit_costs = [3, 3, 1, 3, 3, 1]
##    edit_costs = [7, 3, 5, 9, 2, 6]
#    algo_options = '--threads 8 --initial-solutions 40 --ratio-runs-from-initial-solutions 1'
#    params_ged = {'lib': 'gedlibpy', 'cost': 'CONSTANT', 'method': 'IPFP', 
#                'algo_options': algo_options, 'stabilizer': None, 
#                'edit_cost_constant': edit_costs}    
#    _, ged_mat, _ = compute_geds(Gn, params_ged=params_ged, parallel=True)
#    np.savez('results/test_k_closest_graphs/ged_mat.' + fname_medians + '.with_medians.gm', ged_mat=ged_mat)
    # load from file.
    gmfile = np.load('results/test_k_closest_graphs/ged_mat.' + fname_medians + '.with_medians.gm.npz')
    ged_mat = gmfile['ged_mat']
#    # change medians.
#    edit_costs = [3, 3, 1, 3, 3, 1]
#    algo_options = '--threads 8 --initial-solutions 40 --ratio-runs-from-initial-solutions 1'
#    params_ged = {'lib': 'gedlibpy', 'cost': 'CONSTANT', 'method': 'IPFP', 
#                'algo_options': algo_options, 'stabilizer': None, 
#                'edit_cost_constant': edit_costs}
#    for idx in tqdm(range(len(Gn) - 2), desc='computing GEDs', file=sys.stdout):
#        dis, _, _ = GED(Gn[idx], set_median, **params_ged)
#        ged_mat[idx, -2] = dis
#        ged_mat[-2, idx] = dis
#        dis, _, _ = GED(Gn[idx], gen_median, **params_ged)
#        ged_mat[idx, -1] = dis
#        ged_mat[-1, idx] = dis
#    np.savez('results/test_k_closest_graphs/ged_mat.' + fname_medians + '.with_medians.gm', 
#             ged_mat=ged_mat)

    
    # get indices by classes.
    y_idx = get_same_item_indices(y_all)

    # visualization.
    median_set = [22, 29, 54, 74]
    visualize_graph_dataset('ged', 'tsne', draw_figure, 
                            draw_params={'y_idx': y_idx}, dis_mat=ged_mat,
                            median_set=median_set)
    
###############################################################################
    
    
def visualize_distances_in_kernel_monoterpenoides():
    import os

    ds = {'dataset': '../datasets/monoterpenoides/dataset_10+.ds',
          'graph_dir': os.path.dirname(os.path.realpath(__file__))  + '../../datasets/monoterpenoides/'}  # node/edge symb
    Gn_original, y_all = loadDataset(ds['dataset'])
#    Gn = Gn[0:50]
    
    # compute distance matrix
#    median_set = [22, 29, 54, 74]
    gkernel = 'treeletkernel'
    fit_method = 'expert'
    node_label = 'atom'
    edge_label = 'bond_type'
    ds_name = 'monoterpenoides'
    fname_medians = fit_method + '.' + gkernel
    dir_output = 'results/xp_monoterpenoides/'
    repeat = 0
    
    # get indices by classes.
    y_idx = get_same_item_indices(y_all)
    for i, (y, values) in enumerate(y_idx.items()):
        print('\ny =', y)
        k = len(values)
        
        Gn = [Gn_original[g].copy() for g in values]
        # add set median.
        fname_sm = dir_output + 'medians/' + str(int(y)) + '/set_median.k' + str(int(k)) \
            + '.y' + str(int(y)) + '.repeat' + str(repeat) + '.gxl'
        set_median = loadGXL(fname_sm)
        Gn.append(set_median)
        # add generalized median (estimated pre-image.)
        fname_gm = dir_output + 'medians/' + str(int(y)) + '/gen_median.k' + str(int(k)) \
            + '.y' + str(int(y)) + '.repeat' + str(repeat) + '.gxl'
        gen_median = loadGXL(fname_gm)
        Gn.append(gen_median)
    
        # compute distance matrix
        median_set = range(0, len(values))
    
        Gn_median_set = [Gn[i].copy() for i in median_set]
        Kmatrix_median = compute_kernel(Gn + Gn_median_set, gkernel, node_label, 
                                        edge_label, False)
        Kmatrix = Kmatrix_median[0:len(Gn), 0:len(Gn)]
        dis_mat, _, _, _ = kernel_distance_matrix(Gn, node_label, edge_label, 
                                                  Kmatrix=Kmatrix, gkernel=gkernel)
        print('average distances: ', np.mean(np.mean(dis_mat[0:len(Gn)-2, 0:len(Gn)-2])))
        print('min distances: ', np.min(np.min(dis_mat[0:len(Gn)-2, 0:len(Gn)-2])))
        print('max distances: ', np.max(np.max(dis_mat[0:len(Gn)-2, 0:len(Gn)-2])))

        # add distances for the image of exact median \psi.
        dis_k_median_list = []
        for idx, g in enumerate(Gn):
            dis_k_median_list.append(dis_gstar(idx, range(len(Gn), len(Gn) + len(Gn_median_set)), 
                                               [1 / len(Gn_median_set)] * len(Gn_median_set),
                                               Kmatrix_median, withterm3=False))
        dis_mat_median = np.zeros((len(Gn) + 1, len(Gn) + 1))
        for i in range(len(Gn)):
            for j in range(i, len(Gn)):
                dis_mat_median[i, j] = dis_mat[i, j]
                dis_mat_median[j, i] = dis_mat_median[i, j]
        for i in range(len(Gn)):
            dis_mat_median[i, -1] = dis_k_median_list[i]
            dis_mat_median[-1, i] = dis_k_median_list[i]
            
    
        # visualization.
#    visualize_graph_dataset('graph-kernel', 'tsne', Gn)
#    visualize_graph_dataset('graph-kernel', 'tsne', draw_figure, 
#                            draw_params={'y_idx': y_idx}, dis_mat=dis_mat_median)
        visualize_graph_dataset('graph-kernel', 'tsne', draw_figure, 
                                draw_params={'y_idx': y_idx}, dis_mat=dis_mat_median,
                                median_set=median_set)
        
    
def visualize_distances_in_ged_monoterpenoides():
    from gklearn.preimage.fitDistance import compute_geds
    from gklearn.preimage.ged import GED
    import os
    
    ds = {'dataset': '../datasets/monoterpenoides/dataset_10+.ds',
          'graph_dir': os.path.dirname(os.path.realpath(__file__)) + '../../datasets/monoterpenoides/'}  # node/edge symb
    Gn_original, y_all = loadDataset(ds['dataset'])
#    Gn = Gn[0:50]
    
    # compute distance matrix
#    median_set = [22, 29, 54, 74]
    gkernel = 'treeletkernel'
    fit_method = 'expert'
    ds_name = 'monoterpenoides'
    fname_medians = fit_method + '.' + gkernel
    dir_output = 'results/xp_monoterpenoides/'
    repeat = 0
#    edit_costs = [0.16229209837639536, 0.06612870523413916, 0.04030113378793905, 0.20723547009415202, 0.3338607220394598, 0.27054392518077297]
    edit_costs = [3, 3, 1, 3, 3, 1]
#    edit_costs = [7, 3, 5, 9, 2, 6]
    
    # get indices by classes.
    y_idx = get_same_item_indices(y_all)
    for i, (y, values) in enumerate(y_idx.items()):
        print('\ny =', y)
        k = len(values)
        
        Gn = [Gn_original[g].copy() for g in values]
        # add set median.
        fname_sm = dir_output + 'medians/' + str(int(y)) + '/set_median.k' + str(int(k)) \
            + '.y' + str(int(y)) + '.repeat' + str(repeat) + '.gxl'
        set_median = loadGXL(fname_sm)
        Gn.append(set_median)
        # add generalized median (estimated pre-image.)
        fname_gm = dir_output + 'medians/' + str(int(y)) + '/gen_median.k' + str(int(k)) \
            + '.y' + str(int(y)) + '.repeat' + str(repeat) + '.gxl'
        gen_median = loadGXL(fname_gm)
        Gn.append(gen_median)
    
    
        # compute/load ged matrix.
        # compute.
        algo_options = '--threads 1 --initial-solutions 40 --ratio-runs-from-initial-solutions 1'
        params_ged = {'dataset': ds_name, 'lib': 'gedlibpy', 'cost': 'CONSTANT', 
                      'method': 'IPFP', 'algo_options': algo_options, 
                      'stabilizer': None, 'edit_cost_constant': edit_costs}    
        _, ged_mat, _ = compute_geds(Gn, params_ged=params_ged, parallel=True)
        np.savez(dir_output + 'ged_mat.' + fname_medians + '.y' + str(int(y)) \
            + '.with_medians.gm', ged_mat=ged_mat)
#        # load from file.
#        gmfile = np.load('dir_output + 'ged_mat.' + fname_medians + '.y' + str(int(y)) + '.with_medians.gm.npz')
#        ged_mat = gmfile['ged_mat']
#        # change medians.
#        algo_options = '--threads 1 --initial-solutions 40 --ratio-runs-from-initial-solutions 1'
#        params_ged = {'lib': 'gedlibpy', 'cost': 'CONSTANT', 'method': 'IPFP', 
#                    'algo_options': algo_options, 'stabilizer': None, 
#                    'edit_cost_constant': edit_costs}
#        for idx in tqdm(range(len(Gn) - 2), desc='computing GEDs', file=sys.stdout):
#            dis, _, _ = GED(Gn[idx], set_median, **params_ged)
#            ged_mat[idx, -2] = dis
#            ged_mat[-2, idx] = dis
#            dis, _, _ = GED(Gn[idx], gen_median, **params_ged)
#            ged_mat[idx, -1] = dis
#            ged_mat[-1, idx] = dis
#        np.savez(dir_output + 'ged_mat.' + fname_medians + '.y' + str(int(y)) + '.with_medians.gm', 
#                 ged_mat=ged_mat)

        # visualization.
        median_set = range(0, len(values))
        visualize_graph_dataset('ged', 'tsne', draw_figure, 
                                draw_params={'y_idx': y_idx}, dis_mat=ged_mat,
                                median_set=median_set)
        
        
###############################################################################
    
    
def visualize_distances_in_kernel_letter_h():
    
    ds = {'dataset': 'cpp_ext/data/collections/Letter.xml',
          'graph_dir': os.path.dirname(os.path.realpath(__file__)) + '/cpp_ext/data/datasets/Letter/HIGH/'}  # node/edge symb
    Gn_original, y_all = loadDataset(ds['dataset'], extra_params=ds['graph_dir'])
#    Gn = Gn[0:50]
    
    # compute distance matrix
#    median_set = [22, 29, 54, 74]
    gkernel = 'structuralspkernel'
    fit_method = 'expert'
    node_label = None
    edge_label = None
    ds_name = 'letter-h'
    fname_medians = fit_method + '.' + gkernel
    dir_output = 'results/xp_letter_h/'
    k = 150
    repeat = 0
    
    # get indices by classes.
    y_idx = get_same_item_indices(y_all)
    for i, (y, values) in enumerate(y_idx.items()):
        print('\ny =', y)
        
        Gn = [Gn_original[g].copy() for g in values]
        # add set median.
        fname_sm = dir_output + 'medians/' + y + '/set_median.k' + str(int(k)) \
            + '.y' + y + '.repeat' + str(repeat) + '.gxl'
        set_median = loadGXL(fname_sm)
        Gn.append(set_median)
        # add generalized median (estimated pre-image.)
        fname_gm = dir_output + 'medians/' + y + '/gen_median.k' + str(int(k)) \
            + '.y' + y + '.repeat' + str(repeat) + '.gxl'
        gen_median = loadGXL(fname_gm)
        Gn.append(gen_median)
    
        # compute distance matrix
        median_set = range(0, len(values))
    
        Gn_median_set = [Gn[i].copy() for i in median_set]
        Kmatrix_median = compute_kernel(Gn + Gn_median_set, gkernel, node_label, 
                                        edge_label, False)
        Kmatrix = Kmatrix_median[0:len(Gn), 0:len(Gn)]
        dis_mat, _, _, _ = kernel_distance_matrix(Gn, node_label, edge_label, 
                                                  Kmatrix=Kmatrix, gkernel=gkernel)
        print('average distances: ', np.mean(np.mean(dis_mat[0:len(Gn)-2, 0:len(Gn)-2])))
        print('min distances: ', np.min(np.min(dis_mat[0:len(Gn)-2, 0:len(Gn)-2])))
        print('max distances: ', np.max(np.max(dis_mat[0:len(Gn)-2, 0:len(Gn)-2])))

        # add distances for the image of exact median \psi.
        dis_k_median_list = []
        for idx, g in enumerate(Gn):
            dis_k_median_list.append(dis_gstar(idx, range(len(Gn), len(Gn) + len(Gn_median_set)), 
                                               [1 / len(Gn_median_set)] * len(Gn_median_set),
                                               Kmatrix_median, withterm3=False))
        dis_mat_median = np.zeros((len(Gn) + 1, len(Gn) + 1))
        for i in range(len(Gn)):
            for j in range(i, len(Gn)):
                dis_mat_median[i, j] = dis_mat[i, j]
                dis_mat_median[j, i] = dis_mat_median[i, j]
        for i in range(len(Gn)):
            dis_mat_median[i, -1] = dis_k_median_list[i]
            dis_mat_median[-1, i] = dis_k_median_list[i]
            
    
        # visualization.
#    visualize_graph_dataset('graph-kernel', 'tsne', Gn)
#    visualize_graph_dataset('graph-kernel', 'tsne', draw_figure, 
#                            draw_params={'y_idx': y_idx}, dis_mat=dis_mat_median)
        visualize_graph_dataset('graph-kernel', 'tsne', draw_figure, 
                                draw_params={'y_idx': y_idx}, dis_mat=dis_mat_median,
                                median_set=median_set)
        
    
def visualize_distances_in_ged_letter_h():
    from fitDistance import compute_geds
    from preimage.test_k_closest_graphs import reform_attributes
    
    ds = {'dataset': 'cpp_ext/data/collections/Letter.xml',
          'graph_dir': os.path.dirname(os.path.realpath(__file__)) + '/cpp_ext/data/datasets/Letter/HIGH/'}  # node/edge symb
    Gn_original, y_all = loadDataset(ds['dataset'], extra_params=ds['graph_dir'])
#    Gn = Gn[0:50]
    
    # compute distance matrix
#    median_set = [22, 29, 54, 74]
    gkernel = 'structuralspkernel'
    fit_method = 'expert'
    ds_name = 'letter-h'
    fname_medians = fit_method + '.' + gkernel
    dir_output = 'results/xp_letter_h/'
    k = 150
    repeat = 0
#    edit_costs = [0.16229209837639536, 0.06612870523413916, 0.04030113378793905, 0.20723547009415202, 0.3338607220394598, 0.27054392518077297]
    edit_costs = [3, 3, 1, 3, 3, 1]
#    edit_costs = [7, 3, 5, 9, 2, 6]
    
    # get indices by classes.
    y_idx = get_same_item_indices(y_all)
    for i, (y, values) in enumerate(y_idx.items()):
        print('\ny =', y)
        
        Gn = [Gn_original[g].copy() for g in values]
        # add set median.
        fname_sm = dir_output + 'medians/' + y + '/set_median.k' + str(int(k)) \
            + '.y' + y + '.repeat' + str(repeat) + '.gxl'
        set_median = loadGXL(fname_sm)
        Gn.append(set_median)
        # add generalized median (estimated pre-image.)
        fname_gm = dir_output + 'medians/' + y + '/gen_median.k' + str(int(k)) \
            + '.y' + y + '.repeat' + str(repeat) + '.gxl'
        gen_median = loadGXL(fname_gm)
        Gn.append(gen_median)
    
    
        # compute/load ged matrix.
        # compute.
        algo_options = '--threads 1 --initial-solutions 40 --ratio-runs-from-initial-solutions 1'
        params_ged = {'dataset': 'Letter', 'lib': 'gedlibpy', 'cost': 'CONSTANT', 
                      'method': 'IPFP', 'algo_options': algo_options, 
                      'stabilizer': None, 'edit_cost_constant': edit_costs}    
        for g in Gn:
            reform_attributes(g)
        _, ged_mat, _ = compute_geds(Gn, params_ged=params_ged, parallel=True)
        np.savez(dir_output + 'ged_mat.' + fname_medians + '.y' + y + '.with_medians.gm', ged_mat=ged_mat)
#        # load from file.
#        gmfile = np.load('dir_output + 'ged_mat.' + fname_medians + '.y' + y + '.with_medians.gm.npz')
#        ged_mat = gmfile['ged_mat']
#        # change medians.
#        algo_options = '--threads 1 --initial-solutions 40 --ratio-runs-from-initial-solutions 1'
#        params_ged = {'lib': 'gedlibpy', 'cost': 'CONSTANT', 'method': 'IPFP', 
#                    'algo_options': algo_options, 'stabilizer': None, 
#                    'edit_cost_constant': edit_costs}
#        for idx in tqdm(range(len(Gn) - 2), desc='computing GEDs', file=sys.stdout):
#            dis, _, _ = GED(Gn[idx], set_median, **params_ged)
#            ged_mat[idx, -2] = dis
#            ged_mat[-2, idx] = dis
#            dis, _, _ = GED(Gn[idx], gen_median, **params_ged)
#            ged_mat[idx, -1] = dis
#            ged_mat[-1, idx] = dis
#        np.savez(dir_output + 'ged_mat.' + fname_medians + '.y' + y + '.with_medians.gm', 
#                 ged_mat=ged_mat)

    
        # visualization.
        median_set = range(0, len(values))
        visualize_graph_dataset('ged', 'tsne', draw_figure, 
                                draw_params={'y_idx': y_idx}, dis_mat=ged_mat,
                                median_set=median_set)


if __name__ == '__main__':
    visualize_distances_in_kernel_letter_h()
#    visualize_distances_in_ged_letter_h()
#    visualize_distances_in_kernel_monoterpenoides()
#    visualize_distances_in_kernel_monoterpenoides()
#    visualize_distances_in_kernel()
#    visualize_distances_in_ged()
    
    
    
    
    
    
    
#def draw_figure_dis_k(ax, Gn_embedded, y_idx=None, legend=False):
#    from matplotlib import colors as mcolors
#    colors = list(dict(mcolors.BASE_COLORS, **mcolors.CSS4_COLORS))
##    colors = ['#08306b', '#08519c', '#2171b5', '#4292c6', '#6baed6', '#9ecae1',
##              '#c6dbef', '#deebf7']
#    for i, values in enumerate(y_idx.values()):
#        for item in values:
##            ax.scatter(Gn_embedded[item,0], Gn_embedded[item,1], c=colors[i]) # , c='b')        
#            ax.scatter(Gn_embedded[item,0], Gn_embedded[item,1], c='b')        
#    h1 = ax.scatter(Gn_embedded[[12, 13, 22, 29], 0], Gn_embedded[[12, 13, 22, 29], 1], c='r')
#    h2 = ax.scatter(Gn_embedded[-1, 0], Gn_embedded[-1, 1], c='darkorchid') # \psi
#    h3 = ax.scatter(Gn_embedded[-2, 0], Gn_embedded[-2, 1], c='gold') # gen median
#    h4 = ax.scatter(Gn_embedded[-3, 0], Gn_embedded[-3, 1], c='r', marker='+') # set median
#    if legend:
##    fig.subplots_adjust(bottom=0.17)
#        ax.legend([h1, h2, h3, h4], ['k clostest graphs', 'true median', 'gen median', 'set median'])
##    fig.legend(handles, labels, loc='lower center', ncol=2, frameon=False) # , ncol=5, labelspacing=0.1, handletextpad=0.4, columnspacing=0.6)
##    plt.savefig('symbolic_and_non_comparison_vertical_short.eps', format='eps', dpi=300, transparent=True,
##            bbox_inches='tight')
##    plt.show()
    
    
    
#def draw_figure_ged(ax, Gn_embedded, y_idx=None, legend=False):
#    from matplotlib import colors as mcolors
#    colors = list(dict(mcolors.BASE_COLORS, **mcolors.CSS4_COLORS))
##    colors = ['#08306b', '#08519c', '#2171b5', '#4292c6', '#6baed6', '#9ecae1',
##              '#c6dbef', '#deebf7']
#    for i, values in enumerate(y_idx.values()):
#        for item in values:
##            ax.scatter(Gn_embedded[item,0], Gn_embedded[item,1], c=colors[i]) # , c='b')        
#            ax.scatter(Gn_embedded[item,0], Gn_embedded[item,1], c='b')        
#    h1 = ax.scatter(Gn_embedded[[12, 13, 22, 29], 0], Gn_embedded[[12, 13, 22, 29], 1], c='r')
##    h2 = ax.scatter(Gn_embedded[-1, 0], Gn_embedded[-1, 1], c='darkorchid') # \psi
#    h3 = ax.scatter(Gn_embedded[-1, 0], Gn_embedded[-1, 1], c='gold') # gen median
#    h4 = ax.scatter(Gn_embedded[-2, 0], Gn_embedded[-2, 1], c='r', marker='+') # set median
#    if legend:
##    fig.subplots_adjust(bottom=0.17)
#        ax.legend([h1, h3, h4], ['k clostest graphs', 'gen median', 'set median'])
##    fig.legend(handles, labels, loc='lower center', ncol=2, frameon=False) # , ncol=5, labelspacing=0.1, handletextpad=0.4, columnspacing=0.6)
##    plt.savefig('symbolic_and_non_comparison_vertical_short.eps', format='eps', dpi=300, transparent=True,
##            bbox_inches='tight')
##    plt.show()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 20 16:09:51 2020

@author: ljia

@references:

	[1] S Vichy N Vishwanathan, Nicol N Schraudolph, Risi Kondor, and Karsten M Borgwardt. Graph kernels. Journal of Machine Learning Research, 11(Apr):1201–1242, 2010.
"""

import sys
from gklearn.utils import get_iters
import numpy as np
import networkx as nx
from scipy.sparse import identity
from scipy.sparse.linalg import cg
from gklearn.utils.parallel import parallel_gm, parallel_me
from gklearn.kernels import RandomWalkMeta
from gklearn.utils.utils import compute_vertex_kernels


class ConjugateGradient(RandomWalkMeta):


	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self._node_kernels = kwargs.get('node_kernels', None)
		self._edge_kernels = kwargs.get('edge_kernels', None)
		self._node_labels = kwargs.get('node_labels', [])
		self._edge_labels = kwargs.get('edge_labels', [])
		self._node_attrs = kwargs.get('node_attrs', [])
		self._edge_attrs = kwargs.get('edge_attrs', [])


	def _compute_gm_series(self, graphs):
		self._check_edge_weight(graphs, self.verbose)
		self._check_graphs(graphs)

		lmda = self._weight

		# Compute Gram matrix.
		gram_matrix = np.zeros((len(graphs), len(graphs)))

		# Reindex nodes using consecutive integers for the convenience of kernel computation.
		iterator = get_iters(graphs, desc='Reindex vertices', file=sys.stdout, verbose=(self.verbose >= 2))
		graphs = [nx.convert_node_labels_to_integers(g, first_label=0, label_attribute='label_orignal') for g in iterator]

		if self._p is None and self._q is None: # p and q are uniform distributions as default.

			from itertools import combinations_with_replacement
			itr = combinations_with_replacement(range(0, len(graphs)), 2)
			len_itr = int(len(graphs) * (len(graphs) + 1) / 2)
			iterator = get_iters(itr, desc='Computing kernels', file=sys.stdout, length=len_itr, verbose=(self.verbose >= 2))

			for i, j in iterator:
				kernel = self._kernel_do(graphs[i], graphs[j], lmda)
				gram_matrix[i][j] = kernel
				gram_matrix[j][i] = kernel

		else: # @todo
			pass

		return gram_matrix


	def _compute_gm_imap_unordered(self):
		self._check_edge_weight(self._graphs, self.verbose)
		self._check_graphs(self._graphs)

		# Compute Gram matrix.
		gram_matrix = np.zeros((len(self._graphs), len(self._graphs)))

		# @todo: parallel this.
		# Reindex nodes using consecutive integers for the convenience of kernel computation.
		iterator = get_iters(self._graphs, desc='Reindex vertices', file=sys.stdout, verbose=(self.verbose >= 2))
		self._graphs = [nx.convert_node_labels_to_integers(g, first_label=0, label_attribute='label_orignal') for g in iterator]

		if self._p is None and self._q is None: # p and q are uniform distributions as default.

			def init_worker(gn_toshare):
				global G_gn
				G_gn = gn_toshare

			do_fun = self._wrapper_kernel_do

			parallel_gm(do_fun, gram_matrix, self._graphs, init_worker=init_worker,
						glbv=(self._graphs,), n_jobs=self.n_jobs, verbose=self.verbose)

		else: # @todo
			pass

		return gram_matrix


	def _compute_kernel_list_series(self, g1, g_list):
		self._check_edge_weight(g_list + [g1], self.verbose)
		self._check_graphs(g_list + [g1])

		lmda = self._weight

		# compute kernel list.
		kernel_list = [None] * len(g_list)

		# Reindex nodes using consecutive integers for the convenience of kernel computation.
		g1 = nx.convert_node_labels_to_integers(g1, first_label=0, label_attribute='label_orignal')
		iterator = get_iters(g_list, desc='Reindex vertices', file=sys.stdout, verbose=(self.verbose >= 2))
		g_list = [nx.convert_node_labels_to_integers(g, first_label=0, label_attribute='label_orignal') for g in iterator]

		if self._p is None and self._q is None: # p and q are uniform distributions as default.
			iterator = get_iters(range(len(g_list)), desc='Computing kernels', file=sys.stdout, length=len(g_list), verbose=(self.verbose >= 2))

			for i in iterator:
				kernel = self._kernel_do(g1, g_list[i], lmda)
				kernel_list[i] = kernel

		else: # @todo
			pass

		return kernel_list


	def _compute_kernel_list_imap_unordered(self, g1, g_list):
		self._check_edge_weight(g_list + [g1], self.verbose)
		self._check_graphs(g_list + [g1])

		# compute kernel list.
		kernel_list = [None] * len(g_list)

		# Reindex nodes using consecutive integers for the convenience of kernel computation.
		g1 = nx.convert_node_labels_to_integers(g1, first_label=0, label_attribute='label_orignal')
		# @todo: parallel this.
		iterator = get_iters(g_list, desc='Reindex vertices', file=sys.stdout, verbose=(self.verbose >= 2))
		g_list = [nx.convert_node_labels_to_integers(g, first_label=0, label_attribute='label_orignal') for g in iterator]

		if self._p is None and self._q is None: # p and q are uniform distributions as default.

			def init_worker(g1_toshare, g_list_toshare):
				global G_g1, G_g_list
				G_g1 = g1_toshare
				G_g_list = g_list_toshare

			do_fun = self._wrapper_kernel_list_do

			def func_assign(result, var_to_assign):
				var_to_assign[result[0]] = result[1]
			itr = range(len(g_list))
			len_itr = len(g_list)
			parallel_me(do_fun, func_assign, kernel_list, itr, len_itr=len_itr,
				init_worker=init_worker, glbv=(g1, g_list), method='imap_unordered',
				n_jobs=self.n_jobs, itr_desc='Computing kernels', verbose=self.verbose)

		else: # @todo
			pass

		return kernel_list


	def _wrapper_kernel_list_do(self, itr):
		return itr, self._kernel_do(G_g1, G_g_list[itr], self._weight)


	def _compute_single_kernel_series(self, g1, g2):
		self._check_edge_weight([g1] + [g2], self.verbose)
		self._check_graphs([g1] + [g2])

		lmda = self._weight

		# Reindex nodes using consecutive integers for the convenience of kernel computation.
		g1 = nx.convert_node_labels_to_integers(g1, first_label=0, label_attribute='label_orignal')
		g2 = nx.convert_node_labels_to_integers(g2, first_label=0, label_attribute='label_orignal')

		if self._p is None and self._q is None: # p and q are uniform distributions as default.
			kernel = self._kernel_do(g1, g2, lmda)

		else: # @todo
			pass

		return kernel


	def _kernel_do(self, g1, g2, lmda):

		# Frist, compute kernels between all pairs of nodes using the method borrowed
		# from FCSP. It is faster than directly computing all edge kernels
		# when $d_1d_2>2$, where $d_1$ and $d_2$ are vertex degrees of the
		# graphs compared, which is the most case we went though. For very
		# sparse graphs, this would be slow.
		vk_dict = self._compute_vertex_kernels(g1, g2)

		# Compute the weight matrix of the direct product graph.
		w_times, w_dim = self._compute_weight_matrix(g1, g2, vk_dict)
		# use uniform distribution if there is no prior knowledge.
		p_times_uni = 1 / w_dim
		A = identity(w_times.shape[0]) - w_times * lmda
		b = np.full((w_dim, 1), p_times_uni)
		x, _ = cg(A, b)
		# use uniform distribution if there is no prior knowledge.
		q_times = np.full((1, w_dim), p_times_uni)
		return np.dot(q_times, x)


	def _wrapper_kernel_do(self, itr):
		i = itr[0]
		j = itr[1]
		return i, j, self._kernel_do(G_gn[i], G_gn[j], self._weight)


	def _func_fp(x, p_times, lmda, w_times):
		haha = w_times * x
		haha = lmda * haha
		haha = p_times + haha
		return p_times + lmda * np.dot(w_times, x)


	def _compute_vertex_kernels(self, g1, g2):
		"""Compute vertex kernels between vertices of two graphs.
		"""
		return compute_vertex_kernels(g1, g2, self._node_kernels, node_labels=self._node_labels, node_attrs=self._node_attrs)


	# @todo: move if out to make it faster.
	# @todo: node/edge kernels use direct function rather than dicts.
	def _compute_weight_matrix(self, g1, g2, vk_dict):
		"""Compute the weight matrix of the direct product graph.
		"""
		# Define edge kernels.
		def compute_ek_11(e1, e2, ke):
			e1_labels = [e1[2][el] for el in self._edge_labels]
			e2_labels = [e2[2][el] for el in self._edge_labels]
			e1_attrs = [e1[2][ea] for ea in self._edge_attrs]
			e2_attrs = [e2[2][ea] for ea in self._edge_attrs]
			return ke(e1_labels, e2_labels, e1_attrs, e2_attrs)

		def compute_ek_10(e1, e2, ke):
			e1_labels = [e1[2][el] for el in self._edge_labels]
			e2_labels = [e2[2][el] for el in self._edge_labels]
			return ke(e1_labels, e2_labels)

		def compute_ek_01(e1, e2, ke):
			e1_attrs = [e1[2][ea] for ea in self._edge_attrs]
			e2_attrs = [e2[2][ea] for ea in self._edge_attrs]
			return ke(e1_attrs, e2_attrs)

		def compute_ek_00(e1, e2, ke):
			return 1

		# Select the proper edge kernel.
		if len(self._edge_labels) > 0:
			# edge symb and non-synb labeled
			if len(self._edge_attrs) > 0:
				ke = self._edge_kernels['mix']
				ek_temp = compute_ek_11
			# edge symb labeled
			else:
				ke = self._edge_kernels['symb']
				ek_temp = compute_ek_10
		else:
			# edge non-synb labeled
			if len(self._edge_attrs) > 0:
				ke = self._edge_kernels['nsymb']
				ek_temp = compute_ek_01
			# edge unlabeled
			else:
				ke = None
				ek_temp = compute_ek_00 # @todo: check how much slower is this.

		# Compute the weight matrix.
		w_dim = nx.number_of_nodes(g1) * nx.number_of_nodes(g2)
		w_times = np.zeros((w_dim, w_dim))

		if vk_dict: # node labeled
			if self._ds_infos['directed']:
				for e1 in g1.edges(data=True):
					for e2 in g2.edges(data=True):
						w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0], e1[1] * nx.number_of_nodes(g2) + e2[1])
						w_times[w_idx] = vk_dict[(e1[0], e2[0])] * ek_temp(e1, e2, ke) * vk_dict[(e1[1], e2[1])]
			else: # undirected
				for e1 in g1.edges(data=True):
					for e2 in g2.edges(data=True):
						w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0], e1[1] * nx.number_of_nodes(g2) + e2[1])
						w_times[w_idx] = vk_dict[(e1[0], e2[0])] * ek_temp(e1, e2, ke) * vk_dict[(e1[1], e2[1])] + vk_dict[(e1[0], e2[1])] * ek_temp(e1, e2, ke) * vk_dict[(e1[1], e2[0])]
						w_times[w_idx[1], w_idx[0]] = w_times[w_idx[0], w_idx[1]]
						w_idx2 = (e1[0] * nx.number_of_nodes(g2) + e2[1], e1[1] * nx.number_of_nodes(g2) + e2[0])
						w_times[w_idx2[0], w_idx2[1]] = w_times[w_idx[0], w_idx[1]]
						w_times[w_idx2[1], w_idx2[0]] = w_times[w_idx[0], w_idx[1]]
		else: # node unlabeled
			if self._ds_infos['directed']:
				for e1 in g1.edges(data=True):
					for e2 in g2.edges(data=True):
						w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0], e1[1] * nx.number_of_nodes(g2) + e2[1])
						w_times[w_idx] = ek_temp(e1, e2, ke)
			else: # undirected
				for e1 in g1.edges(data=True):
					for e2 in g2.edges(data=True):
						w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0], e1[1] * nx.number_of_nodes(g2) + e2[1])
						w_times[w_idx] = ek_temp(e1, e2, ke)
						w_times[w_idx[1], w_idx[0]] = w_times[w_idx[0], w_idx[1]]
						w_idx2 = (e1[0] * nx.number_of_nodes(g2) + e2[1], e1[1] * nx.number_of_nodes(g2) + e2[0])
						w_times[w_idx2[0], w_idx2[1]] = w_times[w_idx[0], w_idx[1]]
						w_times[w_idx2[1], w_idx2[0]] = w_times[w_idx[0], w_idx[1]]

		return w_times, w_dim



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr 14 15:16:34 2020

@author: ljia

@references:

	[1] Shervashidze N, Schweitzer P, Leeuwen EJ, Mehlhorn K, Borgwardt KM.
	Weisfeiler-lehman graph kernels. Journal of Machine Learning Research.
	2011;12(Sep):2539-61.
"""

import sys
from collections import Counter
# from functools import partial
from itertools import combinations_with_replacement

import networkx as nx
import numpy as np

from gklearn.kernels import GraphKernel
from gklearn.utils import SpecialLabel
from gklearn.utils.iters import get_iters
from gklearn.utils.parallel import parallel_gm, parallel_me


class WeisfeilerLehman(GraphKernel):  # @todo: sp, edge user kernel.

	def __init__(self, **kwargs):
		GraphKernel.__init__(
			self, **{
				k: kwargs.get(k) for k in
				['parallel', 'n_jobs', 'chunksize', 'normalize',
				 'copy_graphs', 'verbose'] if k in kwargs
			}
		)
		self.node_labels = kwargs.get('node_labels', [])
		self.edge_labels = kwargs.get('edge_labels', [])
		self.height = int(kwargs.get('height', 0))
		self._base_kernel = kwargs.get('base_kernel', 'subtree')
		self._ds_infos = kwargs.get('ds_infos', {})


	##########################################################################
	# The following is the 1st paradigm to compute kernel matrix, which is
	# compatible with `scikit-learn`.
	# -------------------------------------------------------------------
	# Special thanks to the "GraKeL" library for providing an excellent template!
	##########################################################################

	##########################################################################
	# The following is the 2nd paradigm to compute kernel matrix. It is
	# simplified and not compatible with `scikit-learn`.
	##########################################################################

	def _compute_gm_series(self, graphs):
		#		if self.verbose >= 2:
		#			import warnings
		#			warnings.warn('A part of the computation is parallelized.')

		# 		self._add_dummy_node_labels(self._graphs)

		# for WL subtree kernel
		if self._base_kernel == 'subtree':
			gram_matrix = self._subtree_kernel_do(graphs)

		# for WL shortest path kernel
		elif self._base_kernel == 'sp':
			gram_matrix = self._sp_kernel_do(graphs)

		# for WL edge kernel
		elif self._base_kernel == 'edge':
			gram_matrix = self._edge_kernel_do(graphs)

		# for user defined base kernel
		else:
			gram_matrix = self._user_kernel_do(graphs)

		return gram_matrix


	def _compute_gm_imap_unordered(self):
		# 		self._add_dummy_node_labels(self._graphs)

		if self._base_kernel == 'subtree':
			gram_matrix = np.zeros((len(self._graphs), len(self._graphs)))


			#			for i in range(len(self._graphs)):
			#				for j in range(i, len(self._graphs)):
			#					gram_matrix[i][j] = self.pairwise_kernel(self._graphs[i], self._graphs[j])
			#					gram_matrix[j][i] = gram_matrix[i][j]

			def init_worker(gn_toshare):
				global G_gn
				G_gn = gn_toshare


			do_fun = self._wrapper_pairwise
			parallel_gm(
				do_fun, gram_matrix, self._graphs, init_worker=init_worker,
				glbv=(self._graphs,), n_jobs=self.n_jobs, verbose=self.verbose
			)
			return gram_matrix
		else:
			if self.verbose >= 2:
				import warnings
				warnings.warn(
					'This base kernel is not parallelized. The serial computation '
					'is used instead.'
				)
			return self._compute_gm_series(self._graphs)


	def _compute_kernel_list_series(
			self, g1, g_list
	):  # @todo: this should be better.
		#		if self.verbose >= 2:
		#			import warnings
		#			warnings.warn('A part of the computation is parallelized.')

		self._add_dummy_node_labels(g_list + [g1])

		# for WL subtree kernel
		if self._base_kernel == 'subtree':
			gram_matrix = self._subtree_kernel_do(g_list + [g1])

		# for WL shortest path kernel
		elif self._base_kernel == 'sp':
			gram_matrix = self._sp_kernel_do(g_list + [g1])

		# for WL edge kernel
		elif self._base_kernel == 'edge':
			gram_matrix = self._edge_kernel_do(g_list + [g1])

		# for user defined base kernel
		else:
			gram_matrix = self._user_kernel_do(g_list + [g1])

		return list(gram_matrix[-1][0:-1])


	def _compute_kernel_list_imap_unordered(self, g1, g_list):
		self._add_dummy_node_labels(g_list + [g1])

		if self._base_kernel == 'subtree':
			kernel_list = [None] * len(g_list)


			def init_worker(g1_toshare, g_list_toshare):
				global G_g1, G_g_list
				G_g1 = g1_toshare
				G_g_list = g_list_toshare


			do_fun = self._wrapper_kernel_list_do


			def func_assign(result, var_to_assign):
				var_to_assign[result[0]] = result[1]


			itr = range(len(g_list))
			len_itr = len(g_list)
			parallel_me(
				do_fun, func_assign, kernel_list, itr, len_itr=len_itr,
				init_worker=init_worker, glbv=(g1, g_list),
				method='imap_unordered',
				n_jobs=self.n_jobs, itr_desc='Computing kernels',
				verbose=self.verbose
			)
			return kernel_list
		else:
			if self.verbose >= 2:
				import warnings
				warnings.warn(
					'This base kernel is not parallelized. The serial computation '
					'is used instead.'
				)
			return self._compute_kernel_list_series(g1, g_list)


	def _wrapper_kernel_list_do(self, itr):
		return itr, self.pairwise_kernel(G_g1, G_g_list[itr])


	def _compute_single_kernel_series(
			self, g1, g2
	):  # @todo: this should be better.
		self._add_dummy_node_labels([g1] + [g2])

		# for WL subtree kernel
		if self._base_kernel == 'subtree':
			gram_matrix = self._subtree_kernel_do([g1] + [g2])

		# for WL shortest path kernel
		elif self._base_kernel == 'sp':
			gram_matrix = self._sp_kernel_do([g1] + [g2])

		# for WL edge kernel
		elif self._base_kernel == 'edge':
			gram_matrix = self._edge_kernel_do([g1] + [g2])

		# for user defined base kernel
		else:
			gram_matrix = self._user_kernel_do([g1] + [g2])

		return gram_matrix[0][1]


	##########################################################################
	# The following are the methods used by both diagrams.
	##########################################################################

	def validate_parameters(self):
		"""Validate all parameters for the transformer.

		Returns
		-------
		None.

		"""
		super().validate_parameters()
		if len(self.node_labels) == 0:
			if len(self.edge_labels) == 0:
				self._subtree_kernel_do = self._subtree_kernel_do_unlabeled
			else:
				self._subtree_kernel_do = self._subtree_kernel_do_el
		else:
			if len(self.edge_labels) == 0:
				self._subtree_kernel_do = self._subtree_kernel_do_nl
			else:
				self._subtree_kernel_do = self._subtree_kernel_do_labeled


	def pairwise_kernel(self, g1, g2):
		# 		Gn = [g1.copy(), g2.copy()] # @todo: make sure it is a full deep copy. and faster!
		Gn = [g1, g2]
		# for WL subtree kernel
		if self._base_kernel == 'subtree':
			kernel = self._subtree_kernel_do(Gn, return_mat=False)

		# @todo: other subkernels.

		return kernel


	def _wrapper_pairwise(self, itr):
		i = itr[0]
		j = itr[1]
		return i, j, self.pairwise_kernel(G_gn[i], G_gn[j])


	def _compute_kernel_itr(self, kernel, all_num_of_each_label):
		labels = set(
			list(all_num_of_each_label[0].keys()) +
			list(all_num_of_each_label[1].keys())
		)
		vector1 = np.array(
			[(all_num_of_each_label[0][label]
			  if (label in all_num_of_each_label[0].keys()) else 0)
			 for label in labels]
		)
		vector2 = np.array(
			[(all_num_of_each_label[1][label]
			  if (label in all_num_of_each_label[1].keys()) else 0)
			 for label in labels]
		)
		kernel += np.dot(vector1, vector2)
		return kernel


	def _subtree_kernel_do_nl(self, Gn, return_mat=True):
		"""Compute Weisfeiler-Lehman kernels between graphs with node labels.

		Parameters
		----------
		Gn : List of NetworkX graph
			List of graphs between which the kernels are computed.

		Return
		------
		kernel_matrix : Numpy matrix / float
			Kernel matrix, each element of which is the Weisfeiler-Lehman kernel between 2 praphs.
		"""
		kernel_matrix = (np.zeros((len(Gn), len(Gn))) if return_mat else 0)
		gram_itr_fun = (
			self._compute_gram_itr if return_mat else self._compute_kernel_itr)

		# initial for height = 0
		all_num_of_each_label = []  # number of occurence of each label in each graph in this iteration

		# for each graph
		if self.verbose >= 2:
			iterator = get_iters(Gn, desc='Setting all labels into a tuple')
		else:
			iterator = Gn
		for G in iterator:
			# set all labels into a tuple. # @todo: remove this original labels or not?
			for nd, attrs in G.nodes(
					data=True
			):  # @todo: there may be a better way.
				G.nodes[nd]['lt'] = tuple(
					attrs[name] for name in self.node_labels
				)
			# get the set of original labels
			labels_ori = list(nx.get_node_attributes(G, 'lt').values())
			# number of occurence of each label in G
			all_num_of_each_label.append(dict(Counter(labels_ori)))

		# Compute subtree kernel with the 0th iteration and add it to the final kernel.
		kernel_matrix = gram_itr_fun(kernel_matrix, all_num_of_each_label)

		# iterate each height
		for h in range(1, self.height + 1):
			all_set_compressed = {}  # a dictionary mapping original labels to new ones in all graphs in this iteration
			num_of_labels_occured = 0  # number of the set of letters that occur before as node labels at least once in all graphs
			#		all_labels_ori = set() # all unique orignal labels in all graphs in this iteration
			all_num_of_each_label = []  # number of occurence of each label in G

			# @todo: parallel this part.
			# 			if self.verbose >= 2:
			# 				iterator = get_iters(enumerate(Gn), desc='Going through iteration ' + str(h), length=len(Gn))
			# 			else:
			# 				iterator = enumerate(Gn)
			for G in Gn:
				num_of_labels_occured = self._subtree_1graph_nl(
					G, all_set_compressed, all_num_of_each_label,
					num_of_labels_occured
				)

			# Compute subtree kernel with h iterations and add it to the final kernel
			kernel_matrix = gram_itr_fun(kernel_matrix, all_num_of_each_label)

		return kernel_matrix


	def _subtree_kernel_do_el(self, Gn, return_mat=True):
		"""Compute Weisfeiler-Lehman kernels between graphs with edge labels.

		Parameters
		----------
		Gn : List of NetworkX graph
			List of graphs between which the kernels are computed.

		Return
		------
		kernel_matrix : Numpy matrix
			Kernel matrix, each element of which is the Weisfeiler-Lehman kernel between 2 praphs.
		"""
		kernel_matrix = (np.zeros((len(Gn), len(Gn))) if return_mat else 0)
		gram_itr_fun = (
			self._compute_gram_itr if return_mat else self._compute_kernel_itr)

		# initial for height = 0
		all_num_of_each_label = []  # number of occurence of each label in each graph in this iteration

		# Compute subtree kernel with the 0th iteration and add it to the final kernel.
		# if return a kernel matrix:
		if return_mat:
			iterator = combinations_with_replacement(
				range(0, len(kernel_matrix)), 2
			)
			for i, j in iterator:
				kernel_matrix[i][j] += nx.number_of_nodes(
					Gn[i]
				) * nx.number_of_nodes(Gn[j])
				kernel_matrix[j][i] = kernel_matrix[i][j]
		# if return a single kernel between two graphs:
		else:
			kernel_matrix += nx.number_of_nodes(Gn[0]) * nx.number_of_nodes(
				Gn[1]
			)

		# if h >= 1.
		if self.height > 0:
			# Set all edge labels into a tuple. # @todo: remove this original labels or not?
			if self.verbose >= 2:
				iterator = get_iters(Gn, desc='Setting all labels into a tuple')
			else:
				iterator = Gn
			for G in iterator:
				for n1, n2, attrs in G.edges(
						data=True
				):  # @todo: there may be a better way.
					G.edges[(n1, n2)]['lt'] = tuple(
						attrs[name] for name in self.edge_labels
					)

			# When h == 1, compute the kernel.
			all_set_compressed = {}  # a dictionary mapping original labels to new ones in all graphs in this iteration
			num_of_labels_occured = 0  # number of the set of letters that occur before as node labels at least once in all graphs
			all_num_of_each_label = []  # number of occurence of each label in G

			# @todo: parallel this part.
			for G in Gn:
				num_of_labels_occured = self._subtree_1graph_el(
					G, all_set_compressed, all_num_of_each_label,
					num_of_labels_occured
				)

			# Compute subtree kernel with h iterations and add it to the final kernel.
			kernel_matrix = gram_itr_fun(kernel_matrix, all_num_of_each_label)

		# Iterate along heights (>= 2).
		for h in range(2, self.height + 1):
			all_set_compressed = {}  # a dictionary mapping original labels to new ones in all graphs in this iteration
			num_of_labels_occured = 0  # number of the set of letters that occur before as node labels at least once in all graphs
			all_num_of_each_label = []  # number of occurence of each label in G

			# @todo: parallel this part.
			for G in Gn:
				num_of_labels_occured = self._subtree_1graph_nl(
					G, all_set_compressed, all_num_of_each_label,
					num_of_labels_occured
				)

			# Compute subtree kernel with h iterations and add it to the final kernel.
			kernel_matrix = gram_itr_fun(kernel_matrix, all_num_of_each_label)

		return kernel_matrix


	def _subtree_kernel_do_labeled(self, Gn, return_mat=True):
		"""Compute Weisfeiler-Lehman kernels between graphs with both node and
		edge labels.

		Parameters
		----------
		Gn : List of NetworkX graph
			List of graphs between which the kernels are computed.

		Return
		------
		kernel_matrix : Numpy matrix
			Kernel matrix, each element of which is the Weisfeiler-Lehman kernel between 2 praphs.
		"""
		kernel_matrix = (np.zeros((len(Gn), len(Gn))) if return_mat else 0)
		gram_itr_fun = (
			self._compute_gram_itr if return_mat else self._compute_kernel_itr)

		# initial for height = 0
		all_num_of_each_label = []  # number of occurence of each label in each graph in this iteration

		# Set all node labels into a tuple and get # of occurence of each label.
		if self.verbose >= 2:
			iterator = get_iters(
				Gn, desc='Setting all node labels into a tuple'
			)
		else:
			iterator = Gn
		for G in iterator:
			# Set all node labels into a tuple. # @todo: remove this original labels or not?
			for nd, attrs in G.nodes(
					data=True
			):  # @todo: there may be a better way.
				G.nodes[nd]['lt'] = tuple(
					attrs[name] for name in self.node_labels
				)
			# Get the set of original labels.
			labels_ori = list(nx.get_node_attributes(G, 'lt').values())
			# number of occurence of each label in G
			all_num_of_each_label.append(dict(Counter(labels_ori)))

		# Compute subtree kernel with the 0th iteration and add it to the final kernel.
		kernel_matrix = gram_itr_fun(kernel_matrix, all_num_of_each_label)

		# if h >= 1:
		if self.height > 0:
			# Set all edge labels into a tuple. # @todo: remove this original labels or not?
			if self.verbose >= 2:
				iterator = get_iters(
					Gn, desc='Setting all edge labels into a tuple'
				)
			else:
				iterator = Gn
			for G in iterator:
				for n1, n2, attrs in G.edges(
						data=True
				):  # @todo: there may be a better way.
					G.edges[(n1, n2)]['lt'] = tuple(
						attrs[name] for name in self.edge_labels
					)

			# When h == 1, compute the kernel.
			all_set_compressed = {}  # a dictionary mapping original labels to new ones in all graphs in this iteration
			num_of_labels_occured = 0  # number of the set of letters that occur before as node labels at least once in all graphs
			all_num_of_each_label = []  # number of occurence of each label in G

			# @todo: parallel this part.
			for G in Gn:
				num_of_labels_occured = self._subtree_1graph_labeled(
					G, all_set_compressed, all_num_of_each_label,
					num_of_labels_occured
				)

			# Compute subtree kernel with h iterations and add it to the final kernel.
			kernel_matrix = gram_itr_fun(kernel_matrix, all_num_of_each_label)

		# Iterate along heights.
		for h in range(2, self.height + 1):
			all_set_compressed = {}  # a dictionary mapping original labels to new ones in all graphs in this iteration
			num_of_labels_occured = 0  # number of the set of letters that occur before as node labels at least once in all graphs
			all_num_of_each_label = []  # number of occurence of each label in G

			# @todo: parallel this part.
			for G in Gn:
				num_of_labels_occured = self._subtree_1graph_nl(
					G, all_set_compressed, all_num_of_each_label,
					num_of_labels_occured
				)

			# Compute subtree kernel with h iterations and add it to the final kernel.
			kernel_matrix = gram_itr_fun(kernel_matrix, all_num_of_each_label)

		return kernel_matrix


	def _subtree_kernel_do_unlabeled(self, Gn, return_mat=True):
		"""Compute Weisfeiler-Lehman kernels between graphs without labels.

		Parameters
		----------
		Gn : List of NetworkX graph
			List of graphs between which the kernels are computed.

		Return
		------
		kernel_matrix : Numpy matrix
			Kernel matrix, each element of which is the Weisfeiler-Lehman kernel between 2 praphs.
		"""
		kernel_matrix = (np.zeros((len(Gn), len(Gn))) if return_mat else 0)
		gram_itr_fun = (
			self._compute_gram_itr if return_mat else self._compute_kernel_itr)

		# initial for height = 0
		all_num_of_each_label = []  # number of occurence of each label in each graph in this iteration

		# Compute subtree kernel with the 0th iteration and add it to the final kernel.
		# if return a kernel matrix:
		if return_mat:
			iterator = combinations_with_replacement(
				range(0, len(kernel_matrix)), 2
			)
			for i, j in iterator:
				kernel_matrix[i][j] += nx.number_of_nodes(
					Gn[i]
				) * nx.number_of_nodes(Gn[j])
				kernel_matrix[j][i] = kernel_matrix[i][j]
		# if return a single kernel between two graphs:
		else:
			kernel_matrix += nx.number_of_nodes(Gn[0]) * nx.number_of_nodes(
				Gn[1]
			)

		# if h >= 1.
		if self.height > 0:
			# When h == 1, compute the kernel.
			all_set_compressed = {}  # a dictionary mapping original labels to new ones in all graphs in this iteration
			num_of_labels_occured = 0  # number of the set of letters that occur before as node labels at least once in all graphs
			all_num_of_each_label = []  # number of occurence of each label in G

			# @todo: parallel this part.
			for G in Gn:
				num_of_labels_occured = self._subtree_1graph_unlabeled(
					G, all_set_compressed, all_num_of_each_label,
					num_of_labels_occured
				)

			# Compute subtree kernel with h iterations and add it to the final kernel.
			kernel_matrix = gram_itr_fun(kernel_matrix, all_num_of_each_label)

		# Iterate along heights (>= 2).
		for h in range(2, self.height + 1):
			all_set_compressed = {}  # a dictionary mapping original labels to new ones in all graphs in this iteration
			num_of_labels_occured = 0  # number of the set of letters that occur before as node labels at least once in all graphs
			all_num_of_each_label = []  # number of occurence of each label in G

			# @todo: parallel this part.
			for G in Gn:
				num_of_labels_occured = self._subtree_1graph_nl(
					G, all_set_compressed, all_num_of_each_label,
					num_of_labels_occured
				)

			# Compute subtree kernel with h iterations and add it to the final kernel.
			kernel_matrix = gram_itr_fun(kernel_matrix, all_num_of_each_label)

		return kernel_matrix


	def _subtree_1graph_nl(
			self, G, all_set_compressed, all_num_of_each_label,
			num_of_labels_occured
	):
		all_multisets = []
		for node, attrs in G.nodes(data=True):
			# Multiset-label determination.
			multiset = [G.nodes[neighbors]['lt'] for neighbors in G[node]]
			# sorting each multiset
			multiset.sort()
			multiset = [attrs['lt']] + multiset  # add the prefix
			all_multisets.append(tuple(multiset))

		# label compression
		set_unique = list(set(all_multisets))  # set of unique multiset labels
		# a dictionary mapping original labels to new ones.
		set_compressed = {}
		# If a label occured before, assign its former compressed label;
		# otherwise assign the number of labels occured + 1 as the
		# compressed label.
		for value in set_unique:
			if value in all_set_compressed.keys():  # @todo: put keys() function out of for loop?
				set_compressed[value] = all_set_compressed[value]
			else:
				set_compressed[value] = str(
					num_of_labels_occured + 1
				)  # @todo: remove str? and what if num_of_labels_occured is extremely big.
				num_of_labels_occured += 1

		all_set_compressed.update(set_compressed)

		# Relabel nodes.
		for idx, node in enumerate(G.nodes()):
			G.nodes[node]['lt'] = set_compressed[all_multisets[idx]]

		# Get the set of compressed labels.
		labels_comp = list(nx.get_node_attributes(G, 'lt').values())
		all_num_of_each_label.append(dict(Counter(labels_comp)))

		return num_of_labels_occured


	def _subtree_1graph_el(
			self, G, all_set_compressed, all_num_of_each_label,
			num_of_labels_occured
	):
		all_multisets = []
		# 		for node, attrs in G.nodes(data=True):
		for node in G.nodes():
			# Multiset-label determination.
			multiset = [G.edges[(node, neighbors)]['lt'] for neighbors in
			            G[node]]  # @todo: check reference for this.
			# sorting each multiset
			multiset.sort()
			# 			multiset = [attrs['lt']] + multiset # add the prefix
			all_multisets.append(tuple(multiset))

		# label compression
		set_unique = list(set(all_multisets))  # set of unique multiset labels
		# a dictionary mapping original labels to new ones.
		set_compressed = {}
		# If a label occured before, assign its former compressed label;
		# otherwise assign the number of labels occured + 1 as the
		# compressed label.
		for value in set_unique:
			if value in all_set_compressed.keys():  # @todo: put keys() function out of for loop?
				set_compressed[value] = all_set_compressed[value]
			else:
				set_compressed[value] = str(
					num_of_labels_occured + 1
				)  # @todo: remove str?
				num_of_labels_occured += 1

		all_set_compressed.update(set_compressed)

		# Relabel nodes.
		for idx, node in enumerate(G.nodes()):
			G.nodes[node]['lt'] = set_compressed[all_multisets[idx]]

		# Get the set of compressed labels.
		labels_comp = list(
			nx.get_node_attributes(G, 'lt').values()
		)  # @todo: maybe can be faster.
		all_num_of_each_label.append(dict(Counter(labels_comp)))

		return num_of_labels_occured


	def _subtree_1graph_labeled(
			self, G, all_set_compressed, all_num_of_each_label,
			num_of_labels_occured
	):
		all_multisets = []
		for node, attrs in G.nodes(data=True):
			# Multiset-label determination.
			multiset = [tuple(
				(G.edges[(node, neighbors)]['lt'], G.nodes[neighbors]['lt'])
			) for neighbors in G[node]]  # @todo: check reference for this.
			# sorting each multiset
			multiset.sort()
			multiset = [attrs['lt']] + multiset  # add the prefix
			all_multisets.append(tuple(multiset))

		# label compression
		set_unique = list(set(all_multisets))  # set of unique multiset labels
		# a dictionary mapping original labels to new ones.
		set_compressed = {}
		# If a label occured before, assign its former compressed label;
		# otherwise assign the number of labels occured + 1 as the
		# compressed label.
		for value in set_unique:
			if value in all_set_compressed.keys():  # @todo: put keys() function out of for loop?
				set_compressed[value] = all_set_compressed[value]
			else:
				set_compressed[value] = str(
					num_of_labels_occured + 1
				)  # @todo: remove str?
				num_of_labels_occured += 1

		all_set_compressed.update(set_compressed)

		# Relabel nodes.
		for idx, node in enumerate(G.nodes()):
			G.nodes[node]['lt'] = set_compressed[all_multisets[idx]]

		# Get the set of compressed labels.
		labels_comp = list(nx.get_node_attributes(G, 'lt').values())
		all_num_of_each_label.append(dict(Counter(labels_comp)))

		return num_of_labels_occured


	def _subtree_1graph_unlabeled(
			self, G, all_set_compressed, all_num_of_each_label,
			num_of_labels_occured
	):
		# 		all_multisets = []
		# 		for node, attrs in G.nodes(data=True): # @todo: it can be better.
		# 			# Multiset-label determination.
		# 			multiset = [0 for neighbors in G[node]]
		# 			# sorting each multiset
		# 			multiset.sort()
		# 			multiset = [0] + multiset # add the prefix
		# 			all_multisets.append(tuple(multiset))
		all_multisets = [len(G[node]) for node in G.nodes()]

		# label compression
		set_unique = list(set(all_multisets))  # set of unique multiset labels
		# a dictionary mapping original labels to new ones.
		set_compressed = {}
		# If a label occurred before, assign its former compressed label;
		# otherwise assign the number of labels occurred + 1 as the
		# compressed label.
		for value in set_unique:
			if value in all_set_compressed.keys():  # @todo: put keys() function out of for loop?
				set_compressed[value] = all_set_compressed[value]
			else:
				set_compressed[value] = str(
					num_of_labels_occured + 1
				)  # @todo: remove str?
				num_of_labels_occured += 1

		all_set_compressed.update(set_compressed)

		# Relabel nodes.
		for idx, node in enumerate(G.nodes()):
			G.nodes[node]['lt'] = set_compressed[all_multisets[idx]]

		# Get the set of compressed labels.
		labels_comp = list(nx.get_node_attributes(G, 'lt').values())
		all_num_of_each_label.append(dict(Counter(labels_comp)))

		return num_of_labels_occured


	def _compute_gram_itr(self, gram_matrix, all_num_of_each_label):
		"""Compute Gram matrix using the base kernel.
		"""
		#		if self.parallel == 'imap_unordered':
		#			# compute kernels.
		#			def init_worker(alllabels_toshare):
		#				global G_alllabels
		#				G_alllabels = alllabels_toshare
		#			do_partial = partial(self._wrapper_compute_subtree_kernel, gram_matrix)
		#			parallel_gm(do_partial, gram_matrix, Gn, init_worker=init_worker,
		#						glbv=(all_num_of_each_label,), n_jobs=self.n_jobs, verbose=self.verbose)
		#		elif self.parallel is None:
		itr = combinations_with_replacement(range(0, len(gram_matrix)), 2)
		len_itr = int(len(gram_matrix) * (len(gram_matrix) + 1) / 2)
		iterator = get_iters(
			itr, desc='Computing Gram matrix for this iteration',
			file=sys.stdout, length=len_itr, verbose=(self.verbose >= 2)
		)
		for i, j in iterator:
			# 		for i in iterator:
			# 			for j in range(i, len(gram_matrix)):
			gram_matrix[i][j] += self._compute_subtree_kernel(
				all_num_of_each_label[i],
				all_num_of_each_label[j]
			)
			gram_matrix[j][i] = gram_matrix[i][j]

		return gram_matrix


	def _compute_subtree_kernel(self, num_of_each_label1, num_of_each_label2):
		"""Compute the subtree kernel.
		"""
		labels = set(
			list(num_of_each_label1.keys()) + list(num_of_each_label2.keys())
		)
		vector1 = np.array(
			[(num_of_each_label1[label]
			  if (label in num_of_each_label1.keys()) else 0)
			 for label in labels]
		)
		vector2 = np.array(
			[(num_of_each_label2[label]
			  if (label in num_of_each_label2.keys()) else 0)
			 for label in labels]
		)
		kernel = np.dot(vector1, vector2)
		return kernel


	#	def _wrapper_compute_subtree_kernel(self, gram_matrix, itr):
	#		i = itr[0]
	#		j = itr[1]
	#		return i, j, self._compute_subtree_kernel(G_alllabels[i], G_alllabels[j], gram_matrix[i][j])

	def _wl_spkernel_do(Gn, node_label, edge_label, height):
		"""Compute Weisfeiler-Lehman shortest path kernels between graphs.

		Parameters
		----------
		Gn : List of NetworkX graph
			List of graphs between which the kernels are computed.
		node_label : string
			node attribute used as label.
		edge_label : string
			edge attribute used as label.
		height : int
			subtree height.

		Return
		------
		gram_matrix : Numpy matrix
			Kernel matrix, each element of which is the Weisfeiler-Lehman kernel between 2 praphs.
		"""
		pass
		from gklearn.utils.utils import getSPGraph

		# init.
		height = int(height)
		gram_matrix = np.zeros((len(Gn), len(Gn)))  # init kernel

		Gn = [getSPGraph(G, edge_weight=edge_label) for G in
		      Gn]  # get shortest path graphs of Gn

		# initial for height = 0
		for i in range(0, len(Gn)):
			for j in range(i, len(Gn)):
				for e1 in Gn[i].edges(data=True):
					for e2 in Gn[j].edges(data=True):
						if e1[2]['cost'] != 0 and e1[2]['cost'] == e2[2][
							'cost'] and (
								(e1[0] == e2[0] and e1[1] == e2[1]) or (
								e1[0] == e2[1] and e1[1] == e2[0])):
							gram_matrix[i][j] += 1
				gram_matrix[j][i] = gram_matrix[i][j]

		# iterate each height
		for h in range(1, height + 1):
			all_set_compressed = {}  # a dictionary mapping original labels to new ones in all graphs in this iteration
			num_of_labels_occured = 0  # number of the set of letters that occur before as node labels at least once in all graphs
			for G in Gn:  # for each graph
				set_multisets = []
				for node in G.nodes(data=True):
					# Multiset-label determination.
					multiset = [G.node[neighbors][node_label] for neighbors in
					            G[node[0]]]
					# sorting each multiset
					multiset.sort()
					multiset = node[1][node_label] + ''.join(
						multiset
					)  # concatenate to a string and add the prefix
					set_multisets.append(multiset)

				# label compression
				set_unique = list(
					set(set_multisets)
				)  # set of unique multiset labels
				# a dictionary mapping original labels to new ones.
				set_compressed = {}
				# if a label occured before, assign its former compressed label, else assign the number of labels occured + 1 as the compressed label
				for value in set_unique:
					if value in all_set_compressed.keys():
						set_compressed[value] = all_set_compressed[value]
					else:
						set_compressed[value] = str(num_of_labels_occured + 1)
						num_of_labels_occured += 1

				all_set_compressed.update(set_compressed)

				# relabel nodes
				for node in G.nodes(data=True):
					node[1][node_label] = set_compressed[set_multisets[node[0]]]

			# Compute subtree kernel with h iterations and add it to the final kernel
			for i in range(0, len(Gn)):
				for j in range(i, len(Gn)):
					for e1 in Gn[i].edges(data=True):
						for e2 in Gn[j].edges(data=True):
							if e1[2]['cost'] != 0 and e1[2]['cost'] == e2[2][
								'cost'] and (
									(e1[0] == e2[0] and e1[1] == e2[1]) or (
									e1[0] == e2[1] and e1[1] == e2[0])):
								gram_matrix[i][j] += 1
					gram_matrix[j][i] = gram_matrix[i][j]

		return gram_matrix


	def _wl_edgekernel_do(Gn, node_label, edge_label, height):
		"""Compute Weisfeiler-Lehman edge kernels between graphs.

		Parameters
		----------
		Gn : List of NetworkX graph
			List of graphs between which the kernels are computed.
		node_label : string
			node attribute used as label.
		edge_label : string
			edge attribute used as label.
		height : int
			subtree height.

		Return
		------
		gram_matrix : Numpy matrix
			Kernel matrix, each element of which is the Weisfeiler-Lehman kernel between 2 praphs.
		"""
		pass
		# init.
		height = int(height)
		gram_matrix = np.zeros((len(Gn), len(Gn)))  # init kernel

		# initial for height = 0
		for i in range(0, len(Gn)):
			for j in range(i, len(Gn)):
				for e1 in Gn[i].edges(data=True):
					for e2 in Gn[j].edges(data=True):
						if e1[2][edge_label] == e2[2][edge_label] and (
								(e1[0] == e2[0] and e1[1] == e2[1]) or (
								e1[0] == e2[1] and e1[1] == e2[0])):
							gram_matrix[i][j] += 1
				gram_matrix[j][i] = gram_matrix[i][j]

		# iterate each height
		for h in range(1, height + 1):
			all_set_compressed = {}  # a dictionary mapping original labels to new ones in all graphs in this iteration
			num_of_labels_occured = 0  # number of the set of letters that occur before as node labels at least once in all graphs
			for G in Gn:  # for each graph
				set_multisets = []
				for node in G.nodes(data=True):
					# Multiset-label determination.
					multiset = [G.node[neighbors][node_label] for neighbors in
					            G[node[0]]]
					# sorting each multiset
					multiset.sort()
					multiset = node[1][node_label] + ''.join(
						multiset
					)  # concatenate to a string and add the prefix
					set_multisets.append(multiset)

				# label compression
				set_unique = list(
					set(set_multisets)
				)  # set of unique multiset labels
				# a dictionary mapping original labels to new ones.
				set_compressed = {}
				# if a label occured before, assign its former compressed label, else assign the number of labels occured + 1 as the compressed label
				for value in set_unique:
					if value in all_set_compressed.keys():
						set_compressed[value] = all_set_compressed[value]
					else:
						set_compressed[value] = str(num_of_labels_occured + 1)
						num_of_labels_occured += 1

				all_set_compressed.update(set_compressed)

				# relabel nodes
				for node in G.nodes(data=True):
					node[1][node_label] = set_compressed[set_multisets[node[0]]]

			# Compute subtree kernel with h iterations and add it to the final kernel
			for i in range(0, len(Gn)):
				for j in range(i, len(Gn)):
					for e1 in Gn[i].edges(data=True):
						for e2 in Gn[j].edges(data=True):
							if e1[2][edge_label] == e2[2][edge_label] and (
									(e1[0] == e2[0] and e1[1] == e2[1]) or (
									e1[0] == e2[1] and e1[1] == e2[0])):
								gram_matrix[i][j] += 1
					gram_matrix[j][i] = gram_matrix[i][j]

		return gram_matrix


	def _wl_userkernel_do(Gn, node_label, edge_label, height, base_kernel):
		"""Compute Weisfeiler-Lehman kernels based on user-defined kernel between graphs.

		Parameters
		----------
		Gn : List of NetworkX graph
			List of graphs between which the kernels are computed.
		node_label : string
			node attribute used as label.
		edge_label : string
			edge attribute used as label.
		height : int
			subtree height.
		base_kernel : string
			Name of the base kernel function used in each iteration of WL kernel. This function returns a Numpy matrix, each element of which is the user-defined Weisfeiler-Lehman kernel between 2 praphs.

		Return
		------
		gram_matrix : Numpy matrix
			Kernel matrix, each element of which is the Weisfeiler-Lehman kernel between 2 praphs.
		"""
		pass
		# init.
		height = int(height)
		gram_matrix = np.zeros((len(Gn), len(Gn)))  # init kernel

		# initial for height = 0
		gram_matrix = base_kernel(Gn, node_label, edge_label)

		# iterate each height
		for h in range(1, height + 1):
			all_set_compressed = {}  # a dictionary mapping original labels to new ones in all graphs in this iteration
			num_of_labels_occured = 0  # number of the set of letters that occur before as node labels at least once in all graphs
			for G in Gn:  # for each graph
				set_multisets = []
				for node in G.nodes(data=True):
					# Multiset-label determination.
					multiset = [G.node[neighbors][node_label] for neighbors in
					            G[node[0]]]
					# sorting each multiset
					multiset.sort()
					multiset = node[1][node_label] + ''.join(
						multiset
					)  # concatenate to a string and add the prefix
					set_multisets.append(multiset)

				# label compression
				set_unique = list(
					set(set_multisets)
				)  # set of unique multiset labels
				# a dictionary mapping original labels to new ones.
				set_compressed = {}
				# if a label occured before, assign its former compressed label, else assign the number of labels occured + 1 as the compressed label
				for value in set_unique:
					if value in all_set_compressed.keys():
						set_compressed[value] = all_set_compressed[value]
					else:
						set_compressed[value] = str(num_of_labels_occured + 1)
						num_of_labels_occured += 1

				all_set_compressed.update(set_compressed)

				# relabel nodes
				for node in G.nodes(data=True):
					node[1][node_label] = set_compressed[set_multisets[node[0]]]

			# Compute kernel with h iterations and add it to the final kernel
			gram_matrix += base_kernel(Gn, node_label, edge_label)

		return gram_matrix


	def _add_dummy_node_labels(self, Gn):
		if len(self.node_labels) == 0 or (
				len(self.node_labels) == 1 and self.node_labels[
			0] == SpecialLabel.DUMMY):
			for i in range(len(Gn)):
				nx.set_node_attributes(Gn[i], '0', SpecialLabel.DUMMY)
			self.node_labels = [SpecialLabel.DUMMY]


class WLSubtree(WeisfeilerLehman):

	def __init__(self, **kwargs):
		kwargs['base_kernel'] = 'subtree'
		super().__init__(**kwargs)



"""Tests of graph kernels.
"""

import multiprocessing

import pytest


def chooseDataset(ds_name):
	"""Choose dataset according to name.
	"""
	from gklearn.utils.graphfiles import loadDataset

	# no node labels (and no edge labels).
	if ds_name == 'Alkane':
		ds_file = 'datasets/Alkane/dataset.ds'
		ds_y = 'datasets/Alkane/dataset_boiling_point_names.txt'
		Gn, y = loadDataset(ds_file, filename_y=ds_y)
		for G in Gn:
			for node in G.nodes:
				del G.nodes[node]['attributes']
	# node symbolic labels.
	elif ds_name == 'Acyclic':
		ds_file = 'datasets/acyclic/dataset_bps.ds'
		Gn, y = loadDataset(ds_file)
		for G in Gn:
			for node in G.nodes:
				del G.nodes[node]['attributes']
	# node non-symbolic labels.
	elif ds_name == 'Letter-med':
		ds_file = 'datasets/Letter-med/Letter-med_A.txt'
		Gn, y = loadDataset(ds_file)
	# node symbolic and non-symbolic labels (and edge symbolic labels).
	elif ds_name == 'AIDS':
		ds_file = 'datasets/AIDS/AIDS_A.txt'
		Gn, y = loadDataset(ds_file)

	# edge non-symbolic labels (no node labels).
	elif ds_name == 'Fingerprint_edge':
		import networkx as nx
		ds_file = 'datasets/Fingerprint/Fingerprint_A.txt'
		Gn, y = loadDataset(ds_file)
		Gn = [(idx, G) for idx, G in enumerate(Gn) if
		      nx.number_of_edges(G) != 0]
		idx = [G[0] for G in Gn]
		Gn = [G[1] for G in Gn]
		y = [y[i] for i in idx]
		for G in Gn:
			G.graph['node_attrs'] = []
			for node in G.nodes:
				del G.nodes[node]['attributes']
				del G.nodes[node]['x']
				del G.nodes[node]['y']
	# edge non-symbolic labels (and node non-symbolic labels).
	elif ds_name == 'Fingerprint':
		import networkx as nx
		ds_file = 'datasets/Fingerprint/Fingerprint_A.txt'
		Gn, y = loadDataset(ds_file)
		Gn = [(idx, G) for idx, G in enumerate(Gn) if
		      nx.number_of_edges(G) != 0]
		idx = [G[0] for G in Gn]
		Gn = [G[1] for G in Gn]
		y = [y[i] for i in idx]
	# edge symbolic and non-symbolic labels (and node symbolic and non-symbolic labels).
	elif ds_name == 'Cuneiform':
		import networkx as nx
		ds_file = 'datasets/Cuneiform/Cuneiform_A.txt'
		Gn, y = loadDataset(ds_file)

	Gn = Gn[0:3]
	y = y[0:3]

	return Gn, y


@pytest.mark.parametrize('ds_name', ['Alkane', 'AIDS'])
@pytest.mark.parametrize('weight,compute_method', [(0.01, 'geo'), (1, 'exp')])
# @pytest.mark.parametrize('parallel', ['imap_unordered', None])
def test_commonwalkkernel(ds_name, weight, compute_method):
	"""Test common walk kernel.
	"""
	from gklearn.kernels.commonWalkKernel import commonwalkkernel

	Gn, y = chooseDataset(ds_name)

	try:
		Kmatrix, run_time, idx = commonwalkkernel(
			Gn,
			node_label='atom',
			edge_label='bond_type',
			weight=weight,
			compute_method=compute_method,
			#                                             parallel=parallel,
			n_jobs=multiprocessing.cpu_count(),
			verbose=True
		)
	except Exception as exception:
		assert False, exception


@pytest.mark.parametrize('ds_name', ['Alkane', 'AIDS'])
@pytest.mark.parametrize('remove_totters', [True, False])
# @pytest.mark.parametrize('parallel', ['imap_unordered', None])
def test_marginalizedkernel(ds_name, remove_totters):
	"""Test marginalized kernel.
	"""
	from gklearn.kernels.marginalizedKernel import marginalizedkernel

	Gn, y = chooseDataset(ds_name)

	try:
		Kmatrix, run_time = marginalizedkernel(
			Gn,
			node_label='atom',
			edge_label='bond_type',
			p_quit=0.5,
			n_iteration=2,
			remove_totters=remove_totters,
			#                                               parallel=parallel,
			n_jobs=multiprocessing.cpu_count(),
			verbose=True
		)
	except Exception as exception:
		assert False, exception


@pytest.mark.parametrize(
	'compute_method,ds_name,sub_kernel',
	[
		#            ('sylvester', 'Alkane', None),
		#            ('conjugate', 'Alkane', None),
		#            ('conjugate', 'AIDS', None),
		#            ('fp', 'Alkane', None),
		#            ('fp', 'AIDS', None),
		('spectral', 'Alkane', 'exp'),
		('spectral', 'Alkane', 'geo'),
	]
)
# @pytest.mark.parametrize('parallel', ['imap_unordered', None])
def test_randomwalkkernel(ds_name, compute_method, sub_kernel):
	"""Test random walk kernel kernel.
	"""
	from gklearn.kernels.randomWalkKernel import randomwalkkernel
	from gklearn.utils.kernels import deltakernel, gaussiankernel, kernelproduct
	import functools

	Gn, y = chooseDataset(ds_name)

	mixkernel = functools.partial(kernelproduct, deltakernel, gaussiankernel)
	sub_kernels = [
		{'symb': deltakernel, 'nsymb': gaussiankernel, 'mix': mixkernel}]
	try:
		Kmatrix, run_time, idx = randomwalkkernel(
			Gn,
			compute_method=compute_method,
			weight=1e-3,
			p=None,
			q=None,
			edge_weight=None,
			node_kernels=sub_kernels,
			edge_kernels=sub_kernels,
			node_label='atom',
			edge_label='bond_type',
			sub_kernel=sub_kernel,
			#                                                  parallel=parallel,
			n_jobs=multiprocessing.cpu_count(),
			verbose=True
		)
	except Exception as exception:
		assert False, exception


@pytest.mark.parametrize(
	'ds_name', ['Alkane', 'Acyclic', 'Letter-med', 'AIDS', 'Fingerprint']
)
# @pytest.mark.parametrize('parallel', ['imap_unordered', None])
@pytest.mark.parametrize('parallel', ['imap_unordered'])
def test_spkernel(ds_name, parallel):
	"""Test shortest path kernel.
	"""
	from gklearn.kernels.spKernel import spkernel
	from gklearn.utils.kernels import deltakernel, gaussiankernel, kernelproduct
	import functools

	Gn, y = chooseDataset(ds_name)

	mixkernel = functools.partial(kernelproduct, deltakernel, gaussiankernel)
	sub_kernels = {
		'symb': deltakernel, 'nsymb': gaussiankernel, 'mix': mixkernel
	}
	try:
		Kmatrix, run_time, idx = spkernel(
			Gn, node_label='atom',
			node_kernels=sub_kernels,
			parallel=parallel, n_jobs=multiprocessing.cpu_count(), verbose=True
		)
	except Exception as exception:
		assert False, exception


# @pytest.mark.parametrize('ds_name', ['Alkane', 'Acyclic', 'Letter-med', 'AIDS', 'Fingerprint'])
@pytest.mark.parametrize(
	'ds_name', ['Alkane', 'Acyclic', 'Letter-med', 'AIDS', 'Fingerprint',
	            'Fingerprint_edge', 'Cuneiform']
)
@pytest.mark.parametrize('parallel', ['imap_unordered', None])
def test_structuralspkernel(ds_name, parallel):
	"""Test structural shortest path kernel.
	"""
	from gklearn.kernels.structuralspKernel import structuralspkernel
	from gklearn.utils.kernels import deltakernel, gaussiankernel, kernelproduct
	import functools

	Gn, y = chooseDataset(ds_name)

	mixkernel = functools.partial(kernelproduct, deltakernel, gaussiankernel)
	sub_kernels = {
		'symb': deltakernel, 'nsymb': gaussiankernel, 'mix': mixkernel
	}
	try:
		Kmatrix, run_time = structuralspkernel(
			Gn, node_label='atom',
			edge_label='bond_type', node_kernels=sub_kernels,
			edge_kernels=sub_kernels,
			parallel=parallel, n_jobs=multiprocessing.cpu_count(),
			verbose=True
		)
	except Exception as exception:
		assert False, exception


@pytest.mark.parametrize('ds_name', ['Alkane', 'AIDS'])
@pytest.mark.parametrize('parallel', ['imap_unordered', None])
# @pytest.mark.parametrize('k_func', ['MinMax', 'tanimoto', None])
@pytest.mark.parametrize('k_func', ['MinMax', 'tanimoto'])
@pytest.mark.parametrize('compute_method', ['trie', 'naive'])
def test_untilhpathkernel(ds_name, parallel, k_func, compute_method):
	"""Test path kernel up to length $h$.
	"""
	from gklearn.kernels.untilHPathKernel import untilhpathkernel

	Gn, y = chooseDataset(ds_name)

	try:
		Kmatrix, run_time = untilhpathkernel(
			Gn, node_label='atom',
			edge_label='bond_type',
			depth=2, k_func=k_func, compute_method=compute_method,
			parallel=parallel,
			n_jobs=multiprocessing.cpu_count(), verbose=True
		)
	except Exception as exception:
		assert False, exception


@pytest.mark.parametrize('ds_name', ['Alkane', 'AIDS'])
@pytest.mark.parametrize('parallel', ['imap_unordered', None])
def test_treeletkernel(ds_name, parallel):
	"""Test treelet kernel.
	"""
	from gklearn.kernels.treeletKernel import treeletkernel
	from gklearn.utils.kernels import polynomialkernel
	import functools

	Gn, y = chooseDataset(ds_name)

	pkernel = functools.partial(polynomialkernel, d=2, c=1e5)
	try:
		Kmatrix, run_time = treeletkernel(
			Gn,
			sub_kernel=pkernel,
			node_label='atom',
			edge_label='bond_type',
			parallel=parallel,
			n_jobs=multiprocessing.cpu_count(),
			verbose=True
		)
	except Exception as exception:
		assert False, exception


@pytest.mark.parametrize('ds_name', ['Acyclic'])
# @pytest.mark.parametrize('base_kernel', ['subtree', 'sp', 'edge'])
@pytest.mark.parametrize('base_kernel', ['subtree'])
@pytest.mark.parametrize('parallel', ['imap_unordered', None])
def test_weisfeilerlehmankernel(ds_name, parallel, base_kernel):
	"""Test Weisfeiler-Lehman kernel.
	"""
	from gklearn.kernels.weisfeilerLehmanKernel import weisfeilerlehmankernel

	Gn, y = chooseDataset(ds_name)

	try:
		Kmatrix, run_time = weisfeilerlehmankernel(
			Gn,
			node_label='atom',
			edge_label='bond_type',
			height=2,
			base_kernel=base_kernel,
			parallel=parallel,
			n_jobs=multiprocessing.cpu_count(),
			verbose=True
		)
	except Exception as exception:
		assert False, exception


if __name__ == "__main__":
	#    test_spkernel('Alkane', 'imap_unordered')
	test_structuralspkernel('Fingerprint_edge', 'imap_unordered')



"""Tests of GEDEnv.
"""


def test_GEDEnv():
	"""Test GEDEnv.
	"""
	"""**1.   Get dataset.**"""

	from gklearn.utils import Dataset

	# Predefined dataset name, use dataset "MUTAG".
	ds_name = 'MUTAG'

	# Initialize a Dataset.
	dataset = Dataset()
	# Load predefined dataset "MUTAG".
	dataset.load_predefined_dataset(ds_name)
	graph1 = dataset.graphs[0]
	graph2 = dataset.graphs[1]

	"""**2.  Compute graph edit distance.**"""

	try:
		from gklearn.ged.env import GEDEnv

		ged_env = GEDEnv()  # initailize GED environment.
		ged_env.set_edit_cost(
			'CONSTANT',  # GED cost type.
			edit_cost_constants=[3, 3, 1, 3, 3, 1]  # edit costs.
		)
		ged_env.add_nx_graph(graph1, '')  # add graph1
		ged_env.add_nx_graph(graph2, '')  # add graph2
		listID = ged_env.get_all_graph_ids()  # get list IDs of graphs
		ged_env.init(
			init_type='LAZY_WITHOUT_SHUFFLED_COPIES'
		)  # initialize GED environment.
		options = {
			'initialization_method': 'RANDOM',  # or 'NODE', etc.
			'threads': 1  # parallel threads.
		}
		ged_env.set_method(
			'BIPARTITE',  # GED method.
			options  # options for GED method.
		)
		ged_env.init_method()  # initialize GED method.

		ged_env.run_method(listID[0], listID[1])  # run.

		pi_forward = ged_env.get_forward_map(
			listID[0], listID[1]
		)  # forward map.
		pi_backward = ged_env.get_backward_map(
			listID[0], listID[1]
		)  # backward map.
		dis = ged_env.get_upper_bound(
			listID[0], listID[1]
		)  # GED bewteen two graphs.

		import networkx as nx
		assert len(pi_forward) == nx.number_of_nodes(graph1), len(
			pi_backward
		) == nx.number_of_nodes(graph2)

	except Exception as exception:
		assert False, exception


if __name__ == "__main__":
	test_GEDEnv()



"""
@author: linlin

@references: 

	[1] S Vichy N Vishwanathan, Nicol N Schraudolph, Risi Kondor, and Karsten M Borgwardt. Graph kernels. Journal of Machine Learning Research, 11(Apr):1201–1242, 2010.
"""

import time
from functools import partial
from tqdm import tqdm
import sys

import networkx as nx
import numpy as np
from scipy.sparse import identity, kron
from scipy.sparse.linalg import cg
from scipy.optimize import fixed_point

from gklearn.utils.graphdataset import get_dataset_attributes
from gklearn.utils.parallel import parallel_gm

def randomwalkkernel(*args,
					 # params for all method.
					 compute_method=None,
					 weight=1, 
					 p=None, 
					 q=None,
					 edge_weight=None,
					 # params for conjugate and fp method.
					 node_kernels=None, 
					 edge_kernels=None,
					 node_label='atom',
					 edge_label='bond_type',
					 # params for spectral method.
					 sub_kernel=None,										  
					 n_jobs=None,
					 chunksize=None,
					 verbose=True):
	"""Compute random walk graph kernels.

	Parameters
	----------
	Gn : List of NetworkX graph
		List of graphs between which the kernels are computed.
	
	G1, G2 : NetworkX graphs
		Two graphs between which the kernel is computed.

	compute_method : string
		Method used to compute kernel. The Following choices are 
		available:

		'sylvester' - Sylvester equation method.

		'conjugate' - conjugate gradient method.

		'fp' - fixed-point iterations.

		'spectral' - spectral decomposition.

	weight : float
		A constant weight set for random walks of length h.

	p : None
		Initial probability distribution on the unlabeled direct product graph 
		of two graphs. It is set to be uniform over all vertices in the direct 
		product graph.

	q : None
		Stopping probability distribution on the unlabeled direct product graph 
		of two graphs. It is set to be uniform over all vertices in the direct 
		product graph.

	edge_weight : float

		Edge attribute name corresponding to the edge weight.
		
	node_kernels: dict
		A dictionary of kernel functions for nodes, including 3 items: 'symb' 
		for symbolic node labels, 'nsymb' for non-symbolic node labels, 'mix' 
		for both labels. The first 2 functions take two node labels as 
		parameters, and the 'mix' function takes 4 parameters, a symbolic and a
		non-symbolic label for each the two nodes. Each label is in form of 2-D
		dimension array (n_samples, n_features). Each function returns a number
		as the kernel value. Ignored when nodes are unlabeled. This argument
		is designated to conjugate gradient method and fixed-point iterations.

	edge_kernels: dict
		A dictionary of kernel functions for edges, including 3 items: 'symb' 
		for symbolic edge labels, 'nsymb' for non-symbolic edge labels, 'mix' 
		for both labels. The first 2 functions take two edge labels as 
		parameters, and the 'mix' function takes 4 parameters, a symbolic and a
		non-symbolic label for each the two edges. Each label is in form of 2-D
		dimension array (n_samples, n_features). Each function returns a number
		as the kernel value. Ignored when edges are unlabeled. This argument
		is designated to conjugate gradient method and fixed-point iterations.

	node_label: string
		Node attribute used as label. The default node label is atom. This 
		argument is designated to conjugate gradient method and fixed-point 
		iterations.

	edge_label : string
		Edge attribute used as label. The default edge label is bond_type. This 
		argument is designated to conjugate gradient method and fixed-point 
		iterations.
		
	sub_kernel: string
		Method used to compute walk kernel. The Following choices are 
		available:
		'exp' : method based on exponential serials.
		'geo' : method based on geometric serials.
		
	n_jobs: int
		Number of jobs for parallelization. 

	Return
	------
	Kmatrix : Numpy matrix
		Kernel matrix, each element of which is the path kernel up to d between 2 praphs.
	"""
	compute_method = compute_method.lower()
	Gn = args[0] if len(args) == 1 else [args[0], args[1]]
	Gn = [g.copy() for g in Gn]

	eweight = None
	if edge_weight is None:
		if verbose:
			print('\n None edge weight specified. Set all weight to 1.\n')
	else:
		try:
			some_weight = list(
				nx.get_edge_attributes(Gn[0], edge_weight).values())[0]
			if isinstance(some_weight, float) or isinstance(some_weight, int):
				eweight = edge_weight
			else:
				if verbose:
					print('\n Edge weight with name %s is not float or integer. Set all weight to 1.\n'
						  % edge_weight)
		except:
			if verbose:
				print('\n Edge weight with name "%s" is not found in the edge attributes. Set all weight to 1.\n'
					  % edge_weight)

	ds_attrs = get_dataset_attributes(
		Gn,
		attr_names=['node_labeled', 'node_attr_dim', 'edge_labeled',
					'edge_attr_dim', 'is_directed'],
		node_label=node_label,
		edge_label=edge_label)
	
	# remove graphs with no edges, as no walk can be found in their structures, 
	# so the weight matrix between such a graph and itself might be zero.
	len_gn = len(Gn)
	Gn = [(idx, G) for idx, G in enumerate(Gn) if nx.number_of_edges(G) != 0]
	idx = [G[0] for G in Gn]
	Gn = [G[1] for G in Gn]
	if len(Gn) != len_gn:
		if verbose:
			print('\n %d graphs are removed as they don\'t contain edges.\n' %
				  (len_gn - len(Gn)))

	start_time = time.time()
	
#	# get vertex and edge concatenated labels for each graph
#	label_list, d = getLabels(Gn, node_label, edge_label, ds_attrs['is_directed'])
#	gmf = filterGramMatrix(A_wave_list[0], label_list[0], ('C', '0', 'O'), ds_attrs['is_directed'])

	if compute_method == 'sylvester':
		if verbose:
			import warnings
			warnings.warn('All labels are ignored.')
		Kmatrix = _sylvester_equation(Gn, weight, p, q, eweight, n_jobs, chunksize, verbose=verbose)

	elif compute_method == 'conjugate':
		Kmatrix = _conjugate_gradient(Gn, weight, p, q, ds_attrs, node_kernels, 
									  edge_kernels, node_label, edge_label, 
									  eweight, n_jobs, chunksize, verbose=verbose)
		
	elif compute_method == 'fp':
		Kmatrix = _fixed_point(Gn, weight, p, q, ds_attrs, node_kernels, 
							   edge_kernels, node_label, edge_label, 
							   eweight, n_jobs, chunksize, verbose=verbose)

	elif compute_method == 'spectral':
		if verbose:
			import warnings
			warnings.warn('All labels are ignored. Only works for undirected graphs.')
		Kmatrix = _spectral_decomposition(Gn, weight, p, q, sub_kernel, 
										  eweight, n_jobs, chunksize, verbose=verbose)

	elif compute_method == 'kron':
		pass
		for i in range(0, len(Gn)):
			for j in range(i, len(Gn)):
				Kmatrix[i][j] = _randomwalkkernel_kron(Gn[i], Gn[j],
													   node_label, edge_label)
				Kmatrix[j][i] = Kmatrix[i][j]
	else:
		raise Exception(
			'compute method name incorrect. Available methods: "sylvester", "conjugate", "fp", "spectral" and "kron".'
		)

	run_time = time.time() - start_time
	if verbose:
		print("\n --- kernel matrix of random walk kernel of size %d built in %s seconds ---"
			  % (len(Gn), run_time))

	return Kmatrix, run_time, idx


###############################################################################
def _sylvester_equation(Gn, lmda, p, q, eweight, n_jobs, chunksize, verbose=True):
	"""Compute walk graph kernels up to n between 2 graphs using Sylvester method.

	Parameters
	----------
	G1, G2 : NetworkX graph
		Graphs between which the kernel is computed.
	node_label : string
		node attribute used as label.
	edge_label : string
		edge attribute used as label.

	Return
	------
	kernel : float
		Kernel between 2 graphs.
	"""
	Kmatrix = np.zeros((len(Gn), len(Gn)))

	if q is None:
		# don't normalize adjacency matrices if q is a uniform vector. Note
		# A_wave_list actually contains the transposes of the adjacency matrices.
		A_wave_list = [
			nx.adjacency_matrix(G, eweight).todense().transpose() for G in 
			(tqdm(Gn, desc='compute adjacency matrices', file=sys.stdout) if
			 verbose else Gn)
		]
#		# normalized adjacency matrices
#		A_wave_list = []
#		for G in tqdm(Gn, desc='compute adjacency matrices', file=sys.stdout):
#			A_tilde = nx.adjacency_matrix(G, eweight).todense().transpose()   
#			norm = A_tilde.sum(axis=0)
#			norm[norm == 0] = 1
#			A_wave_list.append(A_tilde / norm)
		if p is None: # p is uniform distribution as default.
			def init_worker(Awl_toshare):
				global G_Awl
				G_Awl = Awl_toshare
			do_partial = partial(wrapper_se_do, lmda)   
			parallel_gm(do_partial, Kmatrix, Gn, init_worker=init_worker, 
						glbv=(A_wave_list,), n_jobs=n_jobs, chunksize=chunksize, verbose=verbose)
			
#			pbar = tqdm(
#				total=(1 + len(Gn)) * len(Gn) / 2,
#				desc='Computing kernels',
#				file=sys.stdout)
#			for i in range(0, len(Gn)):
#				for j in range(i, len(Gn)):
#					S = lmda * A_wave_list[j]
#					T_t = A_wave_list[i]
#					# use uniform distribution if there is no prior knowledge.
#					nb_pd = len(A_wave_list[i]) * len(A_wave_list[j])
#					p_times_uni = 1 / nb_pd
#					M0 = np.full((len(A_wave_list[j]), len(A_wave_list[i])), p_times_uni)
#					X = dlyap(S, T_t, M0)
#					X = np.reshape(X, (-1, 1), order='F')
#					# use uniform distribution if there is no prior knowledge.
#					q_times = np.full((1, nb_pd), p_times_uni)
#					Kmatrix[i][j] = np.dot(q_times, X)
#					Kmatrix[j][i] = Kmatrix[i][j]
#					pbar.update(1)

	return Kmatrix


def wrapper_se_do(lmda, itr):
	i = itr[0]
	j = itr[1]
	return i, j, _se_do(G_Awl[i], G_Awl[j], lmda)


def _se_do(A_wave1, A_wave2, lmda):
	from control import dlyap
	S = lmda * A_wave2
	T_t = A_wave1
	# use uniform distribution if there is no prior knowledge.
	nb_pd = len(A_wave1) * len(A_wave2)
	p_times_uni = 1 / nb_pd
	M0 = np.full((len(A_wave2), len(A_wave1)), p_times_uni)
	X = dlyap(S, T_t, M0)
	X = np.reshape(X, (-1, 1), order='F')
	# use uniform distribution if there is no prior knowledge.
	q_times = np.full((1, nb_pd), p_times_uni)
	return np.dot(q_times, X)


###############################################################################
def _conjugate_gradient(Gn, lmda, p, q, ds_attrs, node_kernels, edge_kernels, 
						node_label, edge_label, eweight, n_jobs, chunksize, verbose=True):
	"""Compute walk graph kernels up to n between 2 graphs using conjugate method.

	Parameters
	----------
	G1, G2 : NetworkX graph
		Graphs between which the kernel is computed.
	node_label : string
		node attribute used as label.
	edge_label : string
		edge attribute used as label.

	Return
	------
	kernel : float
		Kernel between 2 graphs.
	"""
	Kmatrix = np.zeros((len(Gn), len(Gn)))
	
#	if not ds_attrs['node_labeled'] and ds_attrs['node_attr_dim'] < 1 and \
#		not ds_attrs['edge_labeled'] and ds_attrs['edge_attr_dim'] < 1:
#		# this is faster from unlabeled graphs. @todo: why?
#		if q is None:
#			# don't normalize adjacency matrices if q is a uniform vector. Note
#			# A_wave_list actually contains the transposes of the adjacency matrices.
#			A_wave_list = [
#				nx.adjacency_matrix(G, eweight).todense().transpose() for G in 
#					tqdm(Gn, desc='compute adjacency matrices', file=sys.stdout)
#			]
#			if p is None: # p is uniform distribution as default.
#				def init_worker(Awl_toshare):
#					global G_Awl
#					G_Awl = Awl_toshare
#				do_partial = partial(wrapper_cg_unlabled_do, lmda)   
#				parallel_gm(do_partial, Kmatrix, Gn, init_worker=init_worker, 
#							glbv=(A_wave_list,), n_jobs=n_jobs)
#	else:  
	# reindex nodes using consecutive integers for convenience of kernel computation.
	Gn = [nx.convert_node_labels_to_integers(
			g, first_label=0, label_attribute='label_orignal') for g in (tqdm(
				Gn, desc='reindex vertices', file=sys.stdout) if verbose else Gn)]
	
	if p is None and q is None: # p and q are uniform distributions as default.
		def init_worker(gn_toshare):
			global G_gn
			G_gn = gn_toshare
		do_partial = partial(wrapper_cg_labeled_do, ds_attrs, node_kernels, 
							 node_label, edge_kernels, edge_label, lmda)   
		parallel_gm(do_partial, Kmatrix, Gn, init_worker=init_worker, 
					glbv=(Gn,), n_jobs=n_jobs, chunksize=chunksize, verbose=verbose)  
			
#			pbar = tqdm(
#				total=(1 + len(Gn)) * len(Gn) / 2,
#				desc='Computing kernels',
#				file=sys.stdout)
#			for i in range(0, len(Gn)):
#				for j in range(i, len(Gn)):
#					result = _cg_labled_do(Gn[i], Gn[j], ds_attrs, node_kernels,
#										   node_label, edge_kernels, edge_label, lmda)
#					Kmatrix[i][j] = result
#					Kmatrix[j][i] = Kmatrix[i][j]
#					pbar.update(1)
	return Kmatrix


def wrapper_cg_unlabled_do(lmda, itr):
	i = itr[0]
	j = itr[1]
	return i, j, _cg_unlabled_do(G_Awl[i], G_Awl[j], lmda)


def _cg_unlabled_do(A_wave1, A_wave2, lmda):
	nb_pd = len(A_wave1) * len(A_wave2)
	p_times_uni = 1 / nb_pd
	w_times = kron(A_wave1, A_wave2).todense()
	A = identity(w_times.shape[0]) - w_times * lmda
	b = np.full((nb_pd, 1), p_times_uni)
	x, _ = cg(A, b)
	# use uniform distribution if there is no prior knowledge.
	q_times = np.full((1, nb_pd), p_times_uni)
	return np.dot(q_times, x)


def wrapper_cg_labeled_do(ds_attrs, node_kernels, node_label, edge_kernels, 
						 edge_label, lmda, itr):
	i = itr[0]
	j = itr[1]
	return i, j, _cg_labeled_do(G_gn[i], G_gn[j], ds_attrs, node_kernels, 
							   node_label, edge_kernels, edge_label, lmda)


def _cg_labeled_do(g1, g2, ds_attrs, node_kernels, node_label, 
				  edge_kernels, edge_label, lmda):
	# Frist, compute kernels between all pairs of nodes using the method borrowed
	# from FCSP. It is faster than directly computing all edge kernels 
	# when $d_1d_2>2$, where $d_1$ and $d_2$ are vertex degrees of the
	# graphs compared, which is the most case we went though. For very 
	# sparse graphs, this would be slow.
	vk_dict = computeVK(g1, g2, ds_attrs, node_kernels, node_label)
						   
	# Compute the weight matrix of the direct product graph.   
	w_times, w_dim = computeW(g1, g2, vk_dict, ds_attrs,
							  edge_kernels, edge_label)															
	# use uniform distribution if there is no prior knowledge.
	p_times_uni = 1 / w_dim
	A = identity(w_times.shape[0]) - w_times * lmda
	b = np.full((w_dim, 1), p_times_uni)
	x, _ = cg(A, b)
	# use uniform distribution if there is no prior knowledge.
	q_times = np.full((1, w_dim), p_times_uni)
	return np.dot(q_times, x)


###############################################################################
def _fixed_point(Gn, lmda, p, q, ds_attrs, node_kernels, edge_kernels, 
						 node_label, edge_label, eweight, n_jobs, chunksize, verbose=True):
	"""Compute walk graph kernels up to n between 2 graphs using Fixed-Point method.

	Parameters
	----------
	G1, G2 : NetworkX graph
		Graphs between which the kernel is computed.
	node_label : string
		node attribute used as label.
	edge_label : string
		edge attribute used as label.

	Return
	------
	kernel : float
		Kernel between 2 graphs.
	"""
	

	Kmatrix = np.zeros((len(Gn), len(Gn)))
	
#	if not ds_attrs['node_labeled'] and ds_attrs['node_attr_dim'] < 1 and \
#		not ds_attrs['edge_labeled'] and ds_attrs['edge_attr_dim'] > 1:
#		# this is faster from unlabeled graphs. @todo: why?
#		if q is None:
#			# don't normalize adjacency matrices if q is a uniform vector. Note
#			# A_wave_list actually contains the transposes of the adjacency matrices.
#			A_wave_list = [
#				nx.adjacency_matrix(G, eweight).todense().transpose() for G in 
#					tqdm(Gn, desc='compute adjacency matrices', file=sys.stdout)
#			]
#			if p is None: # p is uniform distribution as default.
#				pbar = tqdm(
#					total=(1 + len(Gn)) * len(Gn) / 2,
#					desc='Computing kernels',
#					file=sys.stdout)
#				for i in range(0, len(Gn)):
#					for j in range(i, len(Gn)):				   
#						# use uniform distribution if there is no prior knowledge.
#						nb_pd = len(A_wave_list[i]) * len(A_wave_list[j])
#						p_times_uni = 1 / nb_pd
#						w_times = kron(A_wave_list[i], A_wave_list[j]).todense()
#						p_times = np.full((nb_pd, 1), p_times_uni)
#						x = fixed_point(func_fp, p_times, args=(p_times, lmda, w_times))
#						# use uniform distribution if there is no prior knowledge.
#						q_times = np.full((1, nb_pd), p_times_uni)
#						Kmatrix[i][j] = np.dot(q_times, x)
#						Kmatrix[j][i] = Kmatrix[i][j]
#						pbar.update(1)
#	else:  
	# reindex nodes using consecutive integers for the convenience of kernel computation.
	Gn = [nx.convert_node_labels_to_integers(
			g, first_label=0, label_attribute='label_orignal') for g in (tqdm(
				Gn, desc='reindex vertices', file=sys.stdout) if verbose else Gn)]
	
	if p is None and q is None: # p and q are uniform distributions as default.
		def init_worker(gn_toshare):
			global G_gn
			G_gn = gn_toshare
		do_partial = partial(wrapper_fp_labeled_do, ds_attrs, node_kernels, 
							 node_label, edge_kernels, edge_label, lmda)   
		parallel_gm(do_partial, Kmatrix, Gn, init_worker=init_worker, 
					glbv=(Gn,), n_jobs=n_jobs, chunksize=chunksize, verbose=verbose)
	return Kmatrix


def wrapper_fp_labeled_do(ds_attrs, node_kernels, node_label, edge_kernels, 
						 edge_label, lmda, itr):
	i = itr[0]
	j = itr[1]
	return i, j, _fp_labeled_do(G_gn[i], G_gn[j], ds_attrs, node_kernels, 
							   node_label, edge_kernels, edge_label, lmda)


def _fp_labeled_do(g1, g2, ds_attrs, node_kernels, node_label, 
				  edge_kernels, edge_label, lmda):
	# Frist, compute kernels between all pairs of nodes using the method borrowed
	# from FCSP. It is faster than directly computing all edge kernels 
	# when $d_1d_2>2$, where $d_1$ and $d_2$ are vertex degrees of the
	# graphs compared, which is the most case we went though. For very 
	# sparse graphs, this would be slow.
	vk_dict = computeVK(g1, g2, ds_attrs, node_kernels, node_label)
						   
	# Compute weight matrix of the direct product graph.   
	w_times, w_dim = computeW(g1, g2, vk_dict, ds_attrs,
							  edge_kernels, edge_label)															
	# use uniform distribution if there is no prior knowledge.
	p_times_uni = 1 / w_dim
	p_times = np.full((w_dim, 1), p_times_uni)
	x = fixed_point(func_fp, p_times, args=(p_times, lmda, w_times),
					xtol=1e-06, maxiter=1000)
	# use uniform distribution if there is no prior knowledge.
	q_times = np.full((1, w_dim), p_times_uni)
	return np.dot(q_times, x)


def func_fp(x, p_times, lmda, w_times):
	haha = w_times * x
	haha = lmda * haha
	haha = p_times + haha
	return p_times + lmda * np.dot(w_times, x)


###############################################################################
def _spectral_decomposition(Gn, weight, p, q, sub_kernel, eweight, n_jobs, chunksize, verbose=True):
	"""Compute walk graph kernels up to n between 2 unlabeled graphs using 
	spectral decomposition method. Labels will be ignored.

	Parameters
	----------
	G1, G2 : NetworkX graph
		Graphs between which the kernel is computed.
	node_label : string
		node attribute used as label.
	edge_label : string
		edge attribute used as label.

	Return
	------
	kernel : float
		Kernel between 2 graphs.
	"""
	Kmatrix = np.zeros((len(Gn), len(Gn)))

	if q is None:
		# precompute the spectral decomposition of each graph.
		P_list = []
		D_list = []
		for G in (tqdm(Gn, desc='spectral decompose', file=sys.stdout) if 
				  verbose else Gn):
			# don't normalize adjacency matrices if q is a uniform vector. Note
			# A actually is the transpose of the adjacency matrix.
			A = nx.adjacency_matrix(G, eweight).todense().transpose()
			ew, ev = np.linalg.eig(A)
			D_list.append(ew)
			P_list.append(ev)
#		P_inv_list = [p.T for p in P_list] # @todo: also works for directed graphs?

		if p is None: # p is uniform distribution as default.			
			q_T_list = [np.full((1, nx.number_of_nodes(G)), 1 / nx.number_of_nodes(G)) for G in Gn]
#			q_T_list = [q.T for q in q_list]
			def init_worker(q_T_toshare, P_toshare, D_toshare):
				global G_q_T, G_P, G_D 
				G_q_T = q_T_toshare
				G_P = P_toshare
				G_D = D_toshare
			do_partial = partial(wrapper_sd_do, weight, sub_kernel)   
			parallel_gm(do_partial, Kmatrix, Gn, init_worker=init_worker, 
						glbv=(q_T_list, P_list, D_list), n_jobs=n_jobs, 
						chunksize=chunksize, verbose=verbose)
			
			
#			pbar = tqdm(
#				total=(1 + len(Gn)) * len(Gn) / 2,
#				desc='Computing kernels',
#				file=sys.stdout)
#			for i in range(0, len(Gn)):
#				for j in range(i, len(Gn)):
#					result = _sd_do(q_T_list[i], q_T_list[j], P_list[i], P_list[j], 
#									D_list[i], D_list[j], weight, sub_kernel)
#					Kmatrix[i][j] = result
#					Kmatrix[j][i] = Kmatrix[i][j]
#					pbar.update(1)
	return Kmatrix


def wrapper_sd_do(weight, sub_kernel, itr):
	i = itr[0]
	j = itr[1]
	return i, j, _sd_do(G_q_T[i], G_q_T[j], G_P[i], G_P[j], G_D[i], G_D[j], 
						weight, sub_kernel)


def _sd_do(q_T1, q_T2, P1, P2, D1, D2, weight, sub_kernel):	
	# use uniform distribution if there is no prior knowledge.
	kl = kron(np.dot(q_T1, P1), np.dot(q_T2, P2)).todense()
	# @todo: this is not be needed when p = q (kr = kl.T) for undirected graphs
#	kr = kron(np.dot(P_inv_list[i], q_list[i]), np.dot(P_inv_list[j], q_list[j])).todense()
	if sub_kernel == 'exp':
		D_diag = np.array([d1 * d2 for d1 in D1 for d2 in D2])
		kmiddle = np.diag(np.exp(weight * D_diag))
	elif sub_kernel == 'geo':
		D_diag = np.array([d1 * d2 for d1 in D1 for d2 in D2])
		kmiddle = np.diag(weight * D_diag)
		kmiddle = np.identity(len(kmiddle)) - weight * kmiddle
		kmiddle = np.linalg.inv(kmiddle)
	return np.dot(np.dot(kl, kmiddle), kl.T)[0, 0]


###############################################################################
def _randomwalkkernel_kron(G1, G2, node_label, edge_label):
	"""Compute walk graph kernels up to n between 2 graphs using nearest Kronecker product approximation method.

	Parameters
	----------
	G1, G2 : NetworkX graph
		Graphs between which the kernel is computed.
	node_label : string
		node attribute used as label.
	edge_label : string
		edge attribute used as label.

	Return
	------
	kernel : float
		Kernel between 2 graphs.
	"""
	pass


###############################################################################
def getLabels(Gn, node_label, edge_label, directed):
	"""Get symbolic labels of a graph dataset, where vertex labels are dealt
	with by concatenating them to the edge labels of adjacent edges.
	"""
	label_list = []
	label_set = set()
	for g in Gn:
		label_g = {}
		for e in g.edges(data=True):
			nl1 = g.node[e[0]][node_label]
			nl2 = g.node[e[1]][node_label]
			if not directed and nl1 > nl2:
				nl1, nl2 = nl2, nl1
			label = (nl1, e[2][edge_label], nl2)
			label_g[(e[0], e[1])] = label
		label_list.append(label_g)  
	label_set = set([l for lg in label_list for l in lg.values()])
	return label_list, len(label_set)


def filterGramMatrix(gmt, label_dict, label, directed):
	"""Compute (the transpose of) the Gram matrix filtered by a label.
	"""
	gmf = np.zeros(gmt.shape)
	for (n1, n2), l in label_dict.items():
		if l == label:
			gmf[n2, n1] = gmt[n2, n1]
			if not directed:
				gmf[n1, n2] = gmt[n1, n2]
	return gmf


def computeVK(g1, g2, ds_attrs, node_kernels, node_label):
	'''Compute vertex kernels between vertices of two graphs.
	'''
	vk_dict = {}  # shortest path matrices dict
	if ds_attrs['node_labeled']:
		# node symb and non-synb labeled
		if ds_attrs['node_attr_dim'] > 0:
			kn = node_kernels['mix']
			for n1 in g1.nodes(data=True):
				for n2 in g2.nodes(data=True):
					vk_dict[(n1[0], n2[0])] = kn(
						n1[1][node_label], n2[1][node_label],
						n1[1]['attributes'], n2[1]['attributes'])
		# node symb labeled
		else:
			kn = node_kernels['symb']
			for n1 in g1.nodes(data=True):
				for n2 in g2.nodes(data=True):
					vk_dict[(n1[0], n2[0])] = kn(n1[1][node_label],
												 n2[1][node_label])
	else:
		# node non-synb labeled
		if ds_attrs['node_attr_dim'] > 0:
			kn = node_kernels['nsymb']
			for n1 in g1.nodes(data=True):
				for n2 in g2.nodes(data=True):
					vk_dict[(n1[0], n2[0])] = kn(n1[1]['attributes'],
												 n2[1]['attributes'])
		# node unlabeled
		else:
			pass
	return vk_dict


def computeW(g1, g2, vk_dict, ds_attrs, edge_kernels, edge_label):
	"""Compute the weight matrix of the direct product graph.
	"""
	w_dim = nx.number_of_nodes(g1) * nx.number_of_nodes(g2)
	w_times = np.zeros((w_dim, w_dim))
	if vk_dict: # node labeled
		if ds_attrs['is_directed']:
			if ds_attrs['edge_labeled']:
				# edge symb and non-synb labeled
				if ds_attrs['edge_attr_dim'] > 0:
					ke = edge_kernels['mix']
					for e1 in g1.edges(data=True):
						for e2 in g2.edges(data=True):
							ek_temp = ke(e1[2][edge_label], e2[2][edge_label],
										 e1[2]['attributes'], e2[2]['attributes'])
							w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0],
									 e1[1] * nx.number_of_nodes(g2) + e2[1])
							w_times[w_idx] = vk_dict[(e1[0], e2[0])] \
								* ek_temp * vk_dict[(e1[1], e2[1])]
				# edge symb labeled
				else:
					ke = edge_kernels['symb']
					for e1 in g1.edges(data=True):
						for e2 in g2.edges(data=True):
							ek_temp = ke(e1[2][edge_label], e2[2][edge_label])
							w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0],
									 e1[1] * nx.number_of_nodes(g2) + e2[1])
							w_times[w_idx] = vk_dict[(e1[0], e2[0])] \
								* ek_temp * vk_dict[(e1[1], e2[1])]
			else:
				# edge non-synb labeled
				if ds_attrs['edge_attr_dim'] > 0:
					ke = edge_kernels['nsymb']
					for e1 in g1.edges(data=True):
						for e2 in g2.edges(data=True):
							ek_temp = ke(e1[2]['attributes'], e2[2]['attributes'])
							w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0],
									 e1[1] * nx.number_of_nodes(g2) + e2[1])
							w_times[w_idx] = vk_dict[(e1[0], e2[0])] \
								* ek_temp * vk_dict[(e1[1], e2[1])]
				# edge unlabeled
				else:
					for e1 in g1.edges(data=True):
						for e2 in g2.edges(data=True):
							w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0],
									 e1[1] * nx.number_of_nodes(g2) + e2[1])
							w_times[w_idx] = vk_dict[(e1[0], e2[0])] \
								* vk_dict[(e1[1], e2[1])]								
		else: # undirected
			if ds_attrs['edge_labeled']:
				# edge symb and non-synb labeled
				if ds_attrs['edge_attr_dim'] > 0:
					ke = edge_kernels['mix']
					for e1 in g1.edges(data=True):
						for e2 in g2.edges(data=True):
							ek_temp = ke(e1[2][edge_label], e2[2][edge_label],
										 e1[2]['attributes'], e2[2]['attributes'])
							w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0],
									 e1[1] * nx.number_of_nodes(g2) + e2[1])
							w_times[w_idx] = vk_dict[(e1[0], e2[0])] \
								* ek_temp * vk_dict[(e1[1], e2[1])] \
								+ vk_dict[(e1[0], e2[1])] \
								* ek_temp * vk_dict[(e1[1], e2[0])]
							w_times[w_idx[1], w_idx[0]] = w_times[w_idx[0], w_idx[1]]
							w_idx2 = (e1[0] * nx.number_of_nodes(g2) + e2[1],
									 e1[1] * nx.number_of_nodes(g2) + e2[0])
							w_times[w_idx2[0], w_idx2[1]] = w_times[w_idx[0], w_idx[1]]
							w_times[w_idx2[1], w_idx2[0]] = w_times[w_idx[0], w_idx[1]]
				# edge symb labeled
				else:
					ke = edge_kernels['symb']
					for e1 in g1.edges(data=True):
						for e2 in g2.edges(data=True):
							ek_temp = ke(e1[2][edge_label], e2[2][edge_label])
							w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0],
									 e1[1] * nx.number_of_nodes(g2) + e2[1])
							w_times[w_idx] = vk_dict[(e1[0], e2[0])] \
								* ek_temp * vk_dict[(e1[1], e2[1])] \
								+ vk_dict[(e1[0], e2[1])] \
								* ek_temp * vk_dict[(e1[1], e2[0])]
							w_times[w_idx[1], w_idx[0]] = w_times[w_idx[0], w_idx[1]]
							w_idx2 = (e1[0] * nx.number_of_nodes(g2) + e2[1],
									 e1[1] * nx.number_of_nodes(g2) + e2[0])
							w_times[w_idx2[0], w_idx2[1]] = w_times[w_idx[0], w_idx[1]]
							w_times[w_idx2[1], w_idx2[0]] = w_times[w_idx[0], w_idx[1]]
			else:
				# edge non-synb labeled
				if ds_attrs['edge_attr_dim'] > 0:
					ke = edge_kernels['nsymb']
					for e1 in g1.edges(data=True):
						for e2 in g2.edges(data=True):
							ek_temp = ke(e1[2]['attributes'], e2[2]['attributes'])
							w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0],
									 e1[1] * nx.number_of_nodes(g2) + e2[1])
							w_times[w_idx] = vk_dict[(e1[0], e2[0])] \
								* ek_temp * vk_dict[(e1[1], e2[1])] \
								+ vk_dict[(e1[0], e2[1])] \
								* ek_temp * vk_dict[(e1[1], e2[0])]
							w_times[w_idx[1], w_idx[0]] = w_times[w_idx[0], w_idx[1]]
							w_idx2 = (e1[0] * nx.number_of_nodes(g2) + e2[1],
									 e1[1] * nx.number_of_nodes(g2) + e2[0])
							w_times[w_idx2[0], w_idx2[1]] = w_times[w_idx[0], w_idx[1]]
							w_times[w_idx2[1], w_idx2[0]] = w_times[w_idx[0], w_idx[1]]
				# edge unlabeled
				else:
					for e1 in g1.edges(data=True):
						for e2 in g2.edges(data=True):
							w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0],
									 e1[1] * nx.number_of_nodes(g2) + e2[1])
							w_times[w_idx] = vk_dict[(e1[0], e2[0])] \
								* vk_dict[(e1[1], e2[1])] \
								+ vk_dict[(e1[0], e2[1])] \
								* vk_dict[(e1[1], e2[0])]
							w_times[w_idx[1], w_idx[0]] = w_times[w_idx[0], w_idx[1]]
							w_idx2 = (e1[0] * nx.number_of_nodes(g2) + e2[1],
									 e1[1] * nx.number_of_nodes(g2) + e2[0])
							w_times[w_idx2[0], w_idx2[1]] = w_times[w_idx[0], w_idx[1]]
							w_times[w_idx2[1], w_idx2[0]] = w_times[w_idx[0], w_idx[1]]
	else: # node unlabeled
		if ds_attrs['is_directed']:
			if ds_attrs['edge_labeled']:
				# edge symb and non-synb labeled
				if ds_attrs['edge_attr_dim'] > 0:
					ke = edge_kernels['mix']
					for e1 in g1.edges(data=True):
						for e2 in g2.edges(data=True):
							ek_temp = ke(e1[2][edge_label], e2[2][edge_label],
										 e1[2]['attributes'], e2[2]['attributes'])
							w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0],
									 e1[1] * nx.number_of_nodes(g2) + e2[1])
							w_times[w_idx] = ek_temp
				# edge symb labeled
				else:
					ke = edge_kernels['symb']
					for e1 in g1.edges(data=True):
						for e2 in g2.edges(data=True):
							ek_temp = ke(e1[2][edge_label], e2[2][edge_label])
							w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0],
									 e1[1] * nx.number_of_nodes(g2) + e2[1])
							w_times[w_idx] = ek_temp
			else:
				# edge non-synb labeled
				if ds_attrs['edge_attr_dim'] > 0:
					ke = edge_kernels['nsymb']
					for e1 in g1.edges(data=True):
						for e2 in g2.edges(data=True):
							ek_temp = ke(e1[2]['attributes'], e2[2]['attributes'])
							w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0],
									 e1[1] * nx.number_of_nodes(g2) + e2[1])
							w_times[w_idx] = ek_temp
				# edge unlabeled
				else:
					for e1 in g1.edges(data=True):
						for e2 in g2.edges(data=True):
							w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0],
									 e1[1] * nx.number_of_nodes(g2) + e2[1])
							w_times[w_idx] = 1								
		else: # undirected
			if ds_attrs['edge_labeled']:
				# edge symb and non-synb labeled
				if ds_attrs['edge_attr_dim'] > 0:
					ke = edge_kernels['mix']
					for e1 in g1.edges(data=True):
						for e2 in g2.edges(data=True):
							ek_temp = ke(e1[2][edge_label], e2[2][edge_label],
										 e1[2]['attributes'], e2[2]['attributes'])
							w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0],
									 e1[1] * nx.number_of_nodes(g2) + e2[1])
							w_times[w_idx] = ek_temp
							w_times[w_idx[1], w_idx[0]] = w_times[w_idx[0], w_idx[1]]
							w_idx2 = (e1[0] * nx.number_of_nodes(g2) + e2[1],
									 e1[1] * nx.number_of_nodes(g2) + e2[0])
							w_times[w_idx2[0], w_idx2[1]] = w_times[w_idx[0], w_idx[1]]
							w_times[w_idx2[1], w_idx2[0]] = w_times[w_idx[0], w_idx[1]]
				# edge symb labeled
				else:
					ke = edge_kernels['symb']
					for e1 in g1.edges(data=True):
						for e2 in g2.edges(data=True):
							ek_temp = ke(e1[2][edge_label], e2[2][edge_label])
							w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0],
									 e1[1] * nx.number_of_nodes(g2) + e2[1])
							w_times[w_idx] = ek_temp
							w_times[w_idx[1], w_idx[0]] = w_times[w_idx[0], w_idx[1]]
							w_idx2 = (e1[0] * nx.number_of_nodes(g2) + e2[1],
									 e1[1] * nx.number_of_nodes(g2) + e2[0])
							w_times[w_idx2[0], w_idx2[1]] = w_times[w_idx[0], w_idx[1]]
							w_times[w_idx2[1], w_idx2[0]] = w_times[w_idx[0], w_idx[1]]
			else:
				# edge non-synb labeled
				if ds_attrs['edge_attr_dim'] > 0:
					ke = edge_kernels['nsymb']
					for e1 in g1.edges(data=True):
						for e2 in g2.edges(data=True):
							ek_temp = ke(e1[2]['attributes'], e2[2]['attributes'])
							w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0],
									 e1[1] * nx.number_of_nodes(g2) + e2[1])
							w_times[w_idx] = ek_temp
							w_times[w_idx[1], w_idx[0]] = w_times[w_idx[0], w_idx[1]]
							w_idx2 = (e1[0] * nx.number_of_nodes(g2) + e2[1],
									 e1[1] * nx.number_of_nodes(g2) + e2[0])
							w_times[w_idx2[0], w_idx2[1]] = w_times[w_idx[0], w_idx[1]]
							w_times[w_idx2[1], w_idx2[0]] = w_times[w_idx[0], w_idx[1]]
				# edge unlabeled
				else:
					for e1 in g1.edges(data=True):
						for e2 in g2.edges(data=True):
							w_idx = (e1[0] * nx.number_of_nodes(g2) + e2[0],
									 e1[1] * nx.number_of_nodes(g2) + e2[1])
							w_times[w_idx] = 1
							w_times[w_idx[1], w_idx[0]] = w_times[w_idx[0], w_idx[1]]
							w_idx2 = (e1[0] * nx.number_of_nodes(g2) + e2[1],
									 e1[1] * nx.number_of_nodes(g2) + e2[0])
							w_times[w_idx2[0], w_idx2[1]] = w_times[w_idx[0], w_idx[1]]
							w_times[w_idx2[1], w_idx2[0]] = w_times[w_idx[0], w_idx[1]]
	return w_times, w_dim



from gklearn.ged.util.lsape_solver import LSAPESolver
from gklearn.ged.util.util import pairwise_ged, compute_geds, get_nb_edit_operations, ged_options_to_string
from gklearn.ged.util.util import compute_geds_cml, label_costs_to_matrix
from gklearn.ged.util.util2 import pairwise_ged2, get_nb_edit_operations2



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug 19 17:24:46 2020

@author: ljia

@references:

	[1] S Vichy N Vishwanathan, Nicol N Schraudolph, Risi Kondor, and Karsten M Borgwardt. Graph kernels. Journal of Machine Learning Research, 11(Apr):1201–1242, 2010.
"""

import sys
from gklearn.utils import get_iters
import numpy as np
import networkx as nx
from control import dlyap
from gklearn.utils.parallel import parallel_gm, parallel_me
from gklearn.kernels import RandomWalkMeta


class SylvesterEquation(RandomWalkMeta):


	def __init__(self, **kwargs):
		super().__init__(**kwargs)


	def _compute_gm_series(self, graphs):
		self._check_edge_weight(graphs, self.verbose)
		self._check_graphs(graphs)
		if self.verbose >= 2:
			import warnings
			warnings.warn('All labels are ignored.')

		lmda = self._weight

		# compute Gram matrix.
		gram_matrix = np.zeros((len(graphs), len(graphs)))

		if self._q is None:
			# don't normalize adjacency matrices if q is a uniform vector. Note
			# A_wave_list actually contains the transposes of the adjacency matrices.
			iterator = get_iters(graphs, desc='compute adjacency matrices', file=sys.stdout, verbose=(self.verbose >= 2))
			A_wave_list = [nx.adjacency_matrix(G, self._edge_weight).todense().transpose() for G in iterator]
	#		# normalized adjacency matrices
	#		A_wave_list = []
	#		for G in tqdm(Gn, desc='compute adjacency matrices', file=sys.stdout):
	#			A_tilde = nx.adjacency_matrix(G, eweight).todense().transpose()
	#			norm = A_tilde.sum(axis=0)
	#			norm[norm == 0] = 1
	#			A_wave_list.append(A_tilde / norm)

			if self._p is None: # p is uniform distribution as default.
				from itertools import combinations_with_replacement
				itr = combinations_with_replacement(range(0, len(graphs)), 2)
				len_itr = int(len(graphs) * (len(graphs) + 1) / 2)
				iterator = get_iters(itr, desc='Computing kernels', file=sys.stdout, length=len_itr, verbose=(self.verbose >= 2))

				for i, j in iterator:
					kernel = self._kernel_do(A_wave_list[i], A_wave_list[j], lmda)
					gram_matrix[i][j] = kernel
					gram_matrix[j][i] = kernel

			else: # @todo
				pass
		else: # @todo
			pass

		return gram_matrix


	def _compute_gm_imap_unordered(self):
		self._check_edge_weight(self._graphs, self.verbose)
		self._check_graphs(self._graphs)
		if self.verbose >= 2:
			import warnings
			warnings.warn('All labels are ignored.')

		# compute Gram matrix.
		gram_matrix = np.zeros((len(self._graphs), len(self._graphs)))

		if self._q is None:
			# don't normalize adjacency matrices if q is a uniform vector. Note
			# A_wave_list actually contains the transposes of the adjacency matrices.
			iterator = get_iters(self._graphs, desc='compute adjacency matrices', file=sys.stdout, verbose=(self.verbose >= 2))
			A_wave_list = [nx.adjacency_matrix(G, self._edge_weight).todense().transpose() for G in iterator] # @todo: parallel?

			if self._p is None: # p is uniform distribution as default.
				def init_worker(A_wave_list_toshare):
					global G_A_wave_list
					G_A_wave_list = A_wave_list_toshare

				do_fun = self._wrapper_kernel_do

				parallel_gm(do_fun, gram_matrix, self._graphs, init_worker=init_worker,
							glbv=(A_wave_list,), n_jobs=self.n_jobs, verbose=self.verbose)

			else: # @todo
				pass
		else: # @todo
			pass

		return gram_matrix


	def _compute_kernel_list_series(self, g1, g_list):
		self._check_edge_weight(g_list + [g1], self.verbose)
		self._check_graphs(g_list + [g1])
		if self.verbose >= 2:
			import warnings
			warnings.warn('All labels are ignored.')

		lmda = self._weight

		# compute kernel list.
		kernel_list = [None] * len(g_list)

		if self._q is None:
			# don't normalize adjacency matrices if q is a uniform vector. Note
			# A_wave_list actually contains the transposes of the adjacency matrices.
			A_wave_1 = nx.adjacency_matrix(g1, self._edge_weight).todense().transpose()
			iterator = get_iters(g_list, desc='compute adjacency matrices', file=sys.stdout, verbose=(self.verbose >= 2))
			A_wave_list = [nx.adjacency_matrix(G, self._edge_weight).todense().transpose() for G in iterator]

			if self._p is None: # p is uniform distribution as default.
				iterator = get_iters(range(len(g_list)), desc='Computing kernels', file=sys.stdout, length=len(g_list), verbose=(self.verbose >= 2))

				for i in iterator:
					kernel = self._kernel_do(A_wave_1, A_wave_list[i], lmda)
					kernel_list[i] = kernel

			else: # @todo
				pass
		else: # @todo
			pass

		return kernel_list


	def _compute_kernel_list_imap_unordered(self, g1, g_list):
		self._check_edge_weight(g_list + [g1], self.verbose)
		self._check_graphs(g_list + [g1])
		if self.verbose >= 2:
			import warnings
			warnings.warn('All labels are ignored.')

		# compute kernel list.
		kernel_list = [None] * len(g_list)

		if self._q is None:
			# don't normalize adjacency matrices if q is a uniform vector. Note
			# A_wave_list actually contains the transposes of the adjacency matrices.
			A_wave_1 = nx.adjacency_matrix(g1, self._edge_weight).todense().transpose()
			iterator = get_iters(g_list, desc='compute adjacency matrices', file=sys.stdout, verbose=(self.verbose >= 2))
			A_wave_list = [nx.adjacency_matrix(G, self._edge_weight).todense().transpose() for G in iterator] # @todo: parallel?

			if self._p is None: # p is uniform distribution as default.
				def init_worker(A_wave_1_toshare, A_wave_list_toshare):
					global G_A_wave_1, G_A_wave_list
					G_A_wave_1 = A_wave_1_toshare
					G_A_wave_list = A_wave_list_toshare

				do_fun = self._wrapper_kernel_list_do

				def func_assign(result, var_to_assign):
					var_to_assign[result[0]] = result[1]
				itr = range(len(g_list))
				len_itr = len(g_list)
				parallel_me(do_fun, func_assign, kernel_list, itr, len_itr=len_itr,
					init_worker=init_worker, glbv=(A_wave_1, A_wave_list), method='imap_unordered',
					n_jobs=self.n_jobs, itr_desc='Computing kernels', verbose=self.verbose)

			else: # @todo
				pass
		else: # @todo
			pass

		return kernel_list


	def _wrapper_kernel_list_do(self, itr):
		return itr, self._kernel_do(G_A_wave_1, G_A_wave_list[itr], self._weight)


	def _compute_single_kernel_series(self, g1, g2):
		self._check_edge_weight([g1] + [g2], self.verbose)
		self._check_graphs([g1] + [g2])
		if self.verbose >= 2:
			import warnings
			warnings.warn('All labels are ignored.')

		lmda = self._weight

		if self._q is None:
			# don't normalize adjacency matrices if q is a uniform vector. Note
			# A_wave_list actually contains the transposes of the adjacency matrices.
			A_wave_1 = nx.adjacency_matrix(g1, self._edge_weight).todense().transpose()
			A_wave_2 = nx.adjacency_matrix(g2, self._edge_weight).todense().transpose()
			if self._p is None: # p is uniform distribution as default.
				kernel = self._kernel_do(A_wave_1, A_wave_2, lmda)
			else: # @todo
				pass
		else: # @todo
			pass

		return kernel


	def _kernel_do(self, A_wave1, A_wave2, lmda):

		S = lmda * A_wave2
		T_t = A_wave1
		# use uniform distribution if there is no prior knowledge.
		nb_pd = len(A_wave1) * len(A_wave2)
		p_times_uni = 1 / nb_pd
		M0 = np.full((len(A_wave2), len(A_wave1)), p_times_uni)
		X = dlyap(S, T_t, M0)
		X = np.reshape(X, (-1, 1), order='F')
		# use uniform distribution if there is no prior knowledge.
		q_times = np.full((1, nb_pd), p_times_uni)
		return np.dot(q_times, X)


	def _wrapper_kernel_do(self, itr):
		i = itr[0]
		j = itr[1]
		return i, j, self._kernel_do(G_A_wave_list[i], G_A_wave_list[j], self._weight)